import copy
from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.configure.validator import Validator

class WithX:

    def __init__(self, xdict={}):
        self.__xdict = CIStringDict()
        for k,v in xdict.items():
            try:
                self.__xdict[Validator.name(k)] = {"wtype" : xdict[k]["wtype"].strip().upper(), "wvalue" : xdict[k]["wvalue"]}
            except Exception as e:
                raise Exception(f"Invalid WithX entry for name >>{k}<<.")

    def has_locator(self, name):
        return name in self.__xdict

    def format_args(self, name, vargs, kwargs):
        if not self.has_locator(name):
            raise Exception("No WithX locator with name {} found.".format(name))

        # Critical to create a copy
        fmt = copy.deepcopy(self.__xdict[name])
        try:
            if fmt["wtype"] in {'ATTR', 'FATTR'}:
                for k,v in fmt["wvalue"].items():
                    fmt["wvalue"][k] = v.format(**kwargs)
                return fmt["wtype"], fmt["wvalue"]
            else:
                return fmt["wtype"], fmt["wvalue"].format(*vargs, **kwargs)
        except Exception as e:
            from arjuna import Arjuna
            Arjuna.get_logger().error(f"Error in processing withx {name} : {fmt} for vargs {vargs} and kwargs {kwargs}")
            raise

    def format(self, name, loc_obj):
        vargs = None
        kwargs = None
        if type(loc_obj) is str:
            vargs = [loc_obj]
            kwargs = dict()
        elif type(loc_obj) in {list, tuple}:
            vargs = loc_obj
            kwargs = dict()
        elif type(loc_obj) is dict:
            vargs = []
            kwargs = loc_obj
        return self.format_args(name, vargs, kwargs)
        

    


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import os

# Selenium construct. Needs to create wrapper exception and use in dispatcher.
from selenium.common.exceptions import StaleElementReferenceException


from arjuna.tpi.guiauto.base.locatable import Locatable
from arjuna.interact.gui.auto.base.dispatchable import _Dispatchable
from arjuna.tpi.guiauto.widget.element import GuiElement
from arjuna.tpi.engine.asserter import AsserterMixIn
from arjuna.tpi.guiauto.source.multielement import GuiMultiElementSource
from arjuna.tpi.tracker import track

@track("debug")
class _GuiPartialElement(GuiElement):

    def __init__(self, gui, multi_element, index: int, dispatcher_element):
        super().__init__(gui, multi_element._wmd)
        self.__multi_element = multi_element
        self.__index = index
        self._dispatcher = dispatcher_element
        self._load_source_parser()

    def create_dispatcher(self):
        pass

    def _is_partial_element(self):
        return True

    @property
    def index(self):
        return self.__index

class GuiMultiElement(AsserterMixIn, Locatable,_Dispatchable):
    '''
        Represents multiple GuiElements found using a same GuiWidgetLocator.

        Not meant to be directly created. It is created using calls from **Gui** object or **GuiNamespace** object of **Gui**.

        Arguments:
            gui: **Gui** object containing this GuiMultiElement.
            wmd: GuiElementMetaData object.
            elements: (optional) List of GuiElements to populate this object instead of locating it.

        Note:
            It supports an index based retrieval of a contained GuiElement:

                .. code-block:: python

                    multielement[<index>]

            It also supports looping:

                .. code-block:: python

                    for element in multi_element:
                        some element code
    '''
    
    def __init__(self, gui, wmd, elements=None): #, parent=None):
        AsserterMixIn.__init__(self)
        Locatable.__init__(self, gui, wmd) #, parent)
        _Dispatchable.__init__(self)
        if elements:
            self.__elements = elements
            self._size = len(elements)
        else:
            self.__elements = list()   
        self._load_source_parser()

    def _configure_partial_elements(self, elem_config):
        '''
            This method is supposed to be called when multielement identification is completed.
            This is not used for usual multi-element.
            It is used by RadioGroup or DropDown etc which are higher level abstractions that use ME.
        '''
        for instance in self.__elements:
            instance.configure(elem_config)

    def _load_source_parser(self):
        self.__source_parser = GuiMultiElementSource(self.elements)

    def __getitem__(self, index):
        return self.__elements[index]

    @property
    def size(self) -> int:
        '''
            Number of GuiElements in this object.
        '''
        return len(self.__elements)

    @property
    def length(self) -> int:
        '''
            Number of GuiElements in this object.
        '''
        return len(self.__elements)

    def __len__(self) -> int:
        return self.size

    def assert_size(self, size, obj_name, msg=None) -> None:
        '''
            Assert number of elements in this GuiMultiElement.

            Arguments:
                size: Expected size.
                obj_name: Used for making the assertion error message useful.
                msg: (Optional) Any custom message that you want to included in assertion error message.
        '''
        self.asserter.assert_equal(self.size, size, msg="{} should have exactly {} elements, but was found to have {} elements.".format(obj_name, size, self.size, self.asserter._format_msg(msg)))

    def assert_min_size(self, size, obj_name, msg=None) -> None:
        '''
            Assert minimum number of elements in this GuiMultiElement.

            Arguments:
                size: Minimum expected size.
                obj_name: Used for making the assertion error message useful.
                msg: (Optional) Any custom message that you want to included in assertion error message.
        '''
        self.asserter.assert_min(self.size, size, msg="{} should have minimum of {} elements, but was found to have {} elements.".format(obj_name, size, self.size, self.asserter._format_msg(msg)))
    
    def assert_max_size(self, size, obj_name, msg=None) -> None:
        '''
            Assert maximum number of elements in this GuiMultiElement.

            Arguments:
                min_size: Maximum expected size.
                obj_name: Used for making the assertion error message useful.
                msg: (Optional) Any custom message that you want to included in assertion error message.
        '''
        self.asserter.assert_max(self.size, size, msg="{} should have maximum of {} elements, but was found to have {} elements.".format(obj_name, size, self.size, self.asserter._format_msg(msg)))

    def assert_empty(self, obj_name, msg=None) -> None:
        '''
            Assert that this GuiMultiElement has no elements.

            Arguments:
                obj_name: Used for making the assertion error message useful.
                msg: (Optional) Any custom message that you want to included in assertion error message.
        '''
        self.asserter.assert_equal(self.size, 0, msg="{} should be empty, but was found to have {} elements.".format(obj_name, self.size, self.asserter._format_msg(msg)))

    def assert_not_empty(self, obj_name, msg=None) -> None:
        '''
            Assert that this GuiMultiElement has atleast one element.

            Arguments:
                obj_name: Used for making the assertion error message useful.
                msg: (Optional) Any custom message that you want to included in assertion error message.
        '''
        self.asserter.assert_greater(self.size, 0, msg="{} is expected to have atleat 1 element, but was found to be empty.{}".format(obj_name, self.size, self.asserter._format_msg(msg)))

    @size.setter
    def _size(self, count):
        # The logic ignores stale elements 
        for i in range(count):
            try:
                e = _GuiPartialElement(self.gui, self, i, self.dispatcher.get_element_at_index(i))
                self.__elements.append(e)
            except StaleElementReferenceException as e:
                pass

    @property
    def random_element(self) -> GuiElement:
        '''
            A GuiElement chosen at random from GuiElments in this GuiMultiElement.
        '''
        return self[random.randint(0, self.size-1)]

    @property
    def first_element(self) -> GuiElement:
        '''
            First GuiElement in this GuiMultiElement.
        '''
        return self[0]

    @property
    def last_element(self) -> GuiElement:
        '''
            Last GuiElement in this GuiMultiElement.
        '''
        return self[len(self)-1]

    def get_element_at_ordinal(self, ordinal: int) -> GuiElement:
        '''
            Get GuiElement at an ordinal (position) in this GuiMultiElement.

            Ordinals are as per human counting. First element is at ordinal 1.
        '''
        return self.__elements[ordinal-1]

    def get_element_by_visible_text(self, text: str) -> GuiElement:
        '''
            Get GuiElement in this GuiMultiElement whose text is as supplied.

            Args:
                text: Text of this target GuiElement.
        '''
        texts = self.__get_all_texts()
        first_index = self.__find_first_text_index(texts, text)
        return self[first_index]

    def get_element_by_value(self, value: str) -> GuiElement:
        '''
            Get GuiElement in this GuiMultiElement whose value attribute content is as supplied.

            Args:
                value: Value attribute content of this target GuiElement.
        '''
        values = self.__get_all_values()
        first_index = self.__find_first_value_index(values, value)
        return self[first_index]

    def __return_attr_values(self, response):
        if "data" in response and "attrValues" in response["data"]:
            return response["data"]["attrValues"]
        else:
            return None

    def are_selected(self) -> bool:
        '''
            Check if all GuiElements in this GuiMultiElement are in selected state.
        '''
        return [instance.is_selected() for instance in self.__elements]

    # getting index attribute when it does not exist returns value attribute.
    # So, not going the Selenium way. Setu would treat index as computer counting.
    def has_index_selected(self, index: int) -> GuiElement:
        '''
            Check if GuiElement at the given index is in selected state.
        '''
        return self[index].is_selected()

    # Ordinal is human counting
    def has_ordinal_selected(self, ordinal):
        '''
            Check if GuiElement at the given ordinal is in selected state.
            
            Ordinals are as per human counting. First element is at ordinal 1.
        '''
        return self.has_index_selected(ordinal-1)

    def __find_first_match_index(self, in_sequence, to_match):
        try:
            return in_sequence.index(to_match)
        except:
            return -1

    def __get_all_texts(self):
        self._load_source_parser()
        texts = self.__source_parser.texts
        return texts

    def __find_first_text_index(self, texts, text):
        first_index = texts.index(text)
        if first_index == -1:
            raise Exception("No option with {} visible text present in drop down.".format(text))
        return first_index

    def __get_all_values(self):
        self._load_source_parser()
        values = self.__source_parser.values
        return values

    def __find_first_value_index(self, values, value):
        first_index = values.index(value)
        if first_index == -1:
            raise Exception("No option with {} value present in drop down.".format(value))
        return first_index

    @property
    def first_selected_element(self) -> GuiElement:
        '''
            First GuiElement which is in selected state.
        '''
        self._load_source_parser()
        booleans = self.are_selected()
        first_index = None
        try:
            first_index = booleans.index(True)
        except:
            raise Exception("No option in drop down is currenlty selected.")
        else:
            return self[first_index]

    @property
    def source(self) -> GuiMultiElementSource:
        '''
            **GuiSource** for this GuiMultiElement.
        '''
        return self.__source_parser

    @property
    def elements(self):
        '''
            List of GuiElements in this GuiMultiElement
        '''
        return self.__elements

    @property
    def filter(self):
        '''
            GuiMultiElementFilter object for this GuiMultiElement
        '''
        return GuiMultiElementFilter(self)


class GuiMultiElementFilter:
    '''
        Build a new GuiMultiElement by filtering an existing one.

        Arguments:
            gui_multi_element: Existing **GuiMultiElement** object.
    '''

    def __init__(self, gui_multi_element):
        self.__gui = gui_multi_element.gui
        self.__wmd = gui_multi_element._wmd
        self.__elements = gui_multi_element.elements
        self.__filtered_elements = self.__elements

    def build(self) -> GuiMultiElement:
        '''
            Build a new `GuiMultiElement`.
        '''
        me = GuiMultiElement(self.__gui, self.__wmd, elements=self.__filtered_elements)
        self.__filtered_elements = self.__elements
        return me

    def active(self) -> 'self':
        '''
            Choose only active/non-stale `GuiElement` s. 

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        out = list()
        for e in self.__filtered_elements:
            try:
                e._get_html()
                out.append(e)
            except StaleElementReferenceException as e:
                pass
        self.__filtered_elements = out
        return self

    def visible(self) -> 'self':
        '''
            Choose only visible `GuiElement` objects. 

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.is_visible()
        ]
        return self

    def attr(self, name) -> 'self':
        '''
            Choose only `GuiElement` objects which contain the given attribute.

            Arguments:
                name: Name of attribute

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if e.source.is_attr_present(name)
        ]
        return self        

    def attr_value(self, name, value) -> 'self':
        '''
            Choose only `GuiElement` objects which contain the given attribute and its value.

            Arguments:
                name: Name of attribute
                value: Full or partial content of the given attribute

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if value in e.source.get_attr_value(name)
        ]
        return self

    def value(self, value) -> 'self':
        '''
            Choose only `GuiElement` objects whose value attribute matches the value argument provided.

            Arguments:
                value: Full or partial content of value attribute.

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if str(value) in e.source.get_value("value")
        ]
        return self

    def text(self, text) -> 'self':
        '''
            Choose only `GuiElement` objects whose text contains the text provided.

            Arguments:
                text: Full or partial text.

            Returns:
                Current `GuiMultiElementFilter` object
        '''
        self.__filtered_elements = [
            e for e in self.__filtered_elements
            if text in e.text
        ]
        return self     

    


import os
import yaml

from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.core.error import YamlError, YamlUndefinedSectionError

class Yaml:

    def __init__(self, *, name, pydict, file_path=None):
        self.__name = name
        self.__ydict = pydict is not None and pydict or dict()
        self.__sections = tuple(self.__ydict.keys())
        self.__ydict = CIStringDict(self.__ydict)
        self.__file_path = file_path

    @property
    def name(self):
        return self.__name

    @property
    def file_path(self):
        return self.__file_path

    def is_empty(self):
        return not self.__ydict

    def get_section(self, name, *, strict=True):
        val = self.get_value(name, strict=strict)
        if val is not None and type(val) is not dict:
            raise YamlError(f"Section content must be a dictionary. Found content >>{val}<< in {name} section.")
        return Yaml(name=name, pydict=val, file_path=self.file_path)

    def get_value(self, name, *, strict=True, as_yaml_str=False):
        if self.has_section(name):
            if as_yaml_str:
                return yaml.dump(self.__ydict[name])
            else:
                return self.__ydict[name]
        else:
            if strict:
                raise YamlUndefinedSectionError(f"Yaml object does not have a section with the name: {name}")
            else:
                return None

    def as_map(self):
        return self.__ydict

    def has_section(self, name):
        return name in self.__ydict

    @property
    def section_names(self):
        return self.__ydict.keys()

    def validate_sections_present(*section_names, atleast_one=False):
        absent_sections = []
        present_section_names = self.section_names
        for section_name in section_names:
            if section_name not in present_section_names:
                absent_sections.append(section_name)

        if len(absent_sections) == section_names or (len(absent_sections) < len(section_names) and not atleast_one):
            raise YamlUndefinedSectionError(f"Yaml object does not contains mandatory sections: {absent_sections}") 

    @classmethod
    def from_file(cls, *, file_path):
        yaml_name = os.path.basename(file_path).split(".yaml")[0]
        f = open(file_path, "r")
        ydict = yaml.load(f, Loader=yaml.SafeLoader)
        f.close()
        return Yaml(name=yaml_name, pydict=ydict, file_path=file_path)

    @classmethod
    def from_str(cls, *, name, contents):
        return Yaml(name=name, pydict=yaml.safe_load(contents))

    def as_str(self):
        return yaml.dump(self.__ydict)



    

    


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Arjuna

Arjuna is a Python based test automation framework developed by Rahul Verma (www.rahulverma.net). Rahul has implemented smaller variants of features in Arjuna across frameworks and organizations, or given advise around it. However Arjuna being a generic library has the most complete implementation of his ideas, away from project specific contexts and constraints.

Arjuna uses **[pytest](https://docs.pytest.org/en/latest/)** as its recommended test engine. Arjuna also provides its markup for some common use cases on top of pytest. If you like, you can make use of Arjuna with any other test engine or custom frameworks as well.

To import all public names for test automation authors with
from arjuna import *

To run arjuna as a module

    $python -m arjuna <options>

To see which command line options are available:

    $python -m arjuna -h

All classes, functions and names supposed to be used in test automation code are available in Tester Programming Interface: **arjuna.tpi** package.
"""

import pkg_resources
import time
import os
import sys

def __join_paths(*paths):
    return os.path.abspath(os.path.join(*paths))

__root_dir = __join_paths(os.path.dirname(os.path.realpath(__file__)), "..")
__importables_dir = __join_paths(__root_dir, "third_party")

sys.path.insert(0, __importables_dir)
sys.path.insert(0, __root_dir)

class _ArjunFacade:

    def __init__(self):
        self.__version = pkg_resources.require("arjuna")[0].version

    def launch(self, raw_args):
        from arjuna.core.reader.textfile import TextResourceReader
        from arjuna.interface.cli import ArjunaCLI
        reader = TextResourceReader("header.txt")
        print(reader.read().format(version=self.__version))
        reader.close()

        print("Parsing CLI Options...")
        cli = ArjunaCLI(raw_args)
        # Initialize the Arjuna Core as per CLI options
        print("Intializing Arjuna...")
        cli.init()

        print("Executing Arjuna Command...")
        cli.execute()

from arjuna.tpi.engine import Arjuna
from arjuna.tpi.engine.test import *
from arjuna.tpi.engine.relation import *
from arjuna.tpi.engine.resource import *
from arjuna.tpi.tracker import track
from arjuna.tpi.engine.data.markup import *
from arjuna.tpi.engine.data.record import *
from arjuna.tpi.engine.data.generator import Random, DataEntity, Locales

from arjuna.tpi.helper.arjtype import NVPair, NVPairs, Screen, Attr, Point, Offset
from arjuna.tpi.guiauto.meta.locator import GuiWidgetLocator
from arjuna.tpi.guiauto.model.app import GuiApp
from arjuna.tpi.guiauto.model.page import GuiPage
from arjuna.tpi.guiauto.model.section import GuiSection
from arjuna.tpi.guiauto.model.dialog import GuiDialog

from arjuna.tpi.error import *
from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.tpi.helper.audit import HardCoded
from arjuna.tpi.helper.xml import NodeLocator
from arjuna.tpi.helper.arjtype import ProcessedKeyDict
from arjuna.tpi.helper.json import Json

from arjuna.tpi.hook.config import Configurator

from arjuna.tpi.log import *
from arjuna.tpi.magic import C, L, R



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import argparse
from functools import partial

from .validation import *
from arjuna.core.constant import ReportFormat


class Parser:

    def __init__(self):
        self.parser = None

    def get_parser(self):
        return self.parser

class ProjectParser(Parser):

    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-p', '--project-root-dir', metavar="test_project_root_dir", dest="project.root.dir", type=project_dir, help = 'Valid absolute path of an existing Arjuna test project. Needed with python -m arjuna call only. With python arjuna_launcher.py you can skip this argument as the script determines its container Arjuna test project automatically.')

    def process(self, arg_dict):
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)


class NewProjectParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-p', '--project-root-dir', metavar="test_project_root_dir", dest="project.root.dir", type=new_project_dir, help = 'Absolute non-existing Project root directory. Last part of directory represents project name. Name of project should be an Alnum 3-30 length with only lower case letters.')

    def process(self, arg_dict):
        if arg_dict['project.root.dir'] is None:
            print("Fatal Error in CLI processing. You must provide a valid project root directory using -p or --project-dir switch", file=sys.stderr)
            sys.exit(1)

class RunParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-r", "--run-id", dest="run.id", metavar="run_id", type=partial(lname_check, "Run ID"), help = 'Alnum 3-30 length. Only lower case letters.', default="mrun")
        self.parser.add_argument('-o', '--output-format', dest="report.formats", type=report_format, action="append", choices=[i for i in ReportFormat.__members__], help='Output/Report format. Can pass any number of these switches.') # choices=['XML', 'HTML'], 
        self.parser.add_argument('--update', dest="static.rid", action='store_true', help = 'Will result in overwriting of report files. Useful during script development.')
        self.parser.add_argument('--dry-run', dest="dry_run", choices=[i for i in DryRunType.__members__], type=dry_run_type, help='Does a dry run. Tests are not executed. Behavior depends on the type passed as argument. SHOW_TESTS - enumerate tests. SHOW_PLAN - enumerates tests and fixtures. RUN_FIXTURES - Executes setup/teardown fixtures and emuerates tests.')
        self.parser.add_argument('-c', '--ref-conf', dest="ref_conf", metavar="config_name", type=str, help="Reference Configuration object name for this run. Default is 'ref'")
        self.parser.add_argument('-ao', '--arjuna-option', dest="ao",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='Arjuna Option. Can pass any number of these switches.')
        self.parser.add_argument('-uo', '--user-option', dest="uo",
                                 nargs=2,
                                 action='append',
                                 metavar=('option', 'value'),
                                 help='User Option. Can pass any number of these switches.')
        self.parser.add_argument('-dl', '--display-level', dest='log.console.level', type=ustr, choices=[i for i in LoggingLevel.__members__],
                                 help="Minimum message level for display. (choose from 'DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL')", default=LoggingLevel.INFO.name)
        self.parser.add_argument('-ll', '--log-level', dest='log.file.level', type=ustr, choices=[i for i in LoggingLevel.__members__],
                                 help="Minimum message level for log file. (choose from 'DEBUG', 'INFO', 'WARN', 'ERROR', 'FATAL')", default=LoggingLevel.DEBUG.name)

    def process(self, arg_dict):
        pass

class RunDefaultGroupParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)

    def process(self, arg_dict):
        pass

class SessionParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-s", "--session-name", dest="run.session.name", metavar="session_def_name", type=partial(lname_check, "Session Name"), help = 'Name of a defined session in test sessions configuration file in <Project Root>/config/sessions.yaml file.')

    def process(self, arg_dict):
        if arg_dict['run.session.name'] is None:
            print("Fatal Error in CLI processing. You must provide a valid session name using -s or --session-name switch", file=sys.stderr)
            sys.exit(1)

class StageParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-s", "--stage-name", dest="stage_name", metavar="stage_name", type=partial(lname_check, "Stage Name"), help = 'Name of a defined stage in test stages configuration file in <Project Root>/config/stages.yaml file.')

    def process(self, arg_dict):
        if arg_dict['stage_name'] is None:
            print("Fatal Error in CLI processing. You must provide a valid test stage name using -s or --stage-name switch", file=sys.stderr)
            sys.exit(1)

class GroupParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument("-g", "--group-name", dest="group_name", metavar="group_name", type=partial(lname_check, "Group Name"), help = 'Name of a defined group in test groups configuration file in <Project Root>/config/groups.yaml file.')

    def process(self, arg_dict):
        if arg_dict['group_name'] is None:
            print("Fatal Error in CLI processing. You must provide a valid test group name using -g or --group-name switch", file=sys.stderr)
            sys.exit(1)

class PickersParser(Parser):
    def __init__(self):
        super().__init__()
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('-ip', '--include-package', dest="ip", action="append", metavar="package_name_or_regex", default=None, help='Name/regex pattern for including test packages. Can pass any number of these switches.')
        self.parser.add_argument('-ep', '--exclude-package', dest="ep", action="append", metavar="package_name_or_regex", default=None, help='Name/regex pattern for excluding test packages. Can pass any number of these switches.')
        self.parser.add_argument('-im', '--include-module', dest="im", action="append", metavar="module_name_or_regex", default=None, help='Name/regex pattern for including test modules. Can pass any number of these switches.')
        self.parser.add_argument('-em', '--exclude-module', dest="em", action="append", metavar="module_name_or_regex", default=None, help='Name/regex pattern for excluding test modules. Can pass any number of these switches.')
        self.parser.add_argument('-it', '--include-test', dest="it", action="append", metavar="test_name_or_regex", default=None, help='Name/regex pattern for including test functions. Can pass any number of these switches.')
        self.parser.add_argument('-et', '--exclude-test', dest="et", action="append", metavar="test_name_or_regex", default=None, help='Name/regex pattern for excluding test functions. Can pass any number of these switches.')

        self.parser.add_argument('-ir', '--include-rule', dest="ir", action="append", metavar="rule", default=None, help='Include test functions that match the rule. Can pass any number of these switches. Test Function is included if any of the inclusion rules matches.')
        self.parser.add_argument('-er', '--exclude-rule', dest="er", action="append", metavar="rule", default=None, help='Exclude test functions that match the rule. Can pass any number of these switches. Test Function is excluded if any of the exclusion rules matches. Evaluated before any inclusion rules.')

    def process(self, arg_dict):
        pass


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from arjuna.core.error import WaitableError, _GuiWidgetPresentError
from arjuna.tpi.error import GuiWidgetNotFoundError

class ElementFinder:
    def __init__(self, container): #, obj_name=""):
        self.__container = container
        self.__config = container.config
        # self.__obj_name = obj_name

    @property
    def config(self):
        return self.__config

    @property
    def container(self):
        return self.__container

    @abc.abstractmethod
    def _create_element_flat_or_nested(self, wmd):
        pass

    @abc.abstractmethod
    def _create_multielement_flat_or_nested(self, wmd):
        pass

    def check_for_absence(self, dispatcher_call, wmd, context="ELEMENT"):
        try:
            self.find(dispatcher_call, wmd, context)
        except GuiWidgetNotFoundError as e:
            # This is expected
            pass
        else:
            raise _GuiWidgetPresentError(*wmd.locators)

    def find(self, dispatcher_call, wmd, context="ELEMENT"):
        from arjuna import Arjuna
        Arjuna.get_logger().debug("Finding with wmd: {}".format(str(wmd)))
        from arjuna import Arjuna
        found = False
        js_call_name = context == "ELEMENT" and "_find_element_with_js" or "_find_multielement_with_js"
        js_call = getattr(self.container, js_call_name)
        locators = wmd.locators
        if context != "ELEMENT":
            if "POINT" in {l.ltype.name for l in locators}:
                raise ConditionException("With.POINT can be used only with GuiElement.")

        we = None
        for locator in locators:
            try:
                if locator.ltype.name == "POINT":
                    # Assumption here is that this container is automator.
                    size, dispatcher = js_call(locator.lvalue)
                elif locator.ltype.name == "JS":
                    size, dispatcher = js_call(locator.lvalue)
                else:
                    lvalue = locator.lvalue
                    if locator.ltype.name == "XPATH":
                        if not lvalue.startswith("."):
                            lvalue = "." + lvalue
                    size, dispatcher = dispatcher_call(locator.ltype.name, lvalue)
                return locator.ltype.name, locator.lvalue, size, dispatcher
            except WaitableError as e:
                we = e
            except Exception as f:
                raise f
            else:
                we = None
        if not found:
            raise GuiWidgetNotFoundError(*wmd.locators, container=self.__container)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

class Asserter:
    '''
        Arjuna's asserter class.

        It can be used directly. It is already included as an attribute for request fixture in test functions and all Guis and Gui Templates.
    '''

    def __init__(self):
        self.__asserter = unittest.TestCase('__init__')

    @classmethod
    def _format_msg(cls, msg):
        return msg and " {}.".format(msg) or ""

    def assert_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 == obj2

            Wrapper on unittest's assertEqual

            Args:
                obj1: Object of any type which supports == operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertEqual(obj1, obj2, msg)

    def assert_lesser(self, obj1, obj2, msg):
        '''
            Assert obj1 < obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports < operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLess(obj1, obj2, msg)

    def assert_greater(self, obj1, obj2, msg):
        '''
            Assert obj1 > obj2

            Wrapper on unittest's assertLess

            Args:
                obj1: Object of any type which supports > operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreater(obj1, obj2, msg)

    def assert_min(self, obj, min_value, msg):
        '''
            Asserts a minimum value for an object i.e. obj >= min_value

            Wrapper on unittest's assertGreaterEqual

            Args:
                obj: Object of any type which supports >= operator for min_value
                min_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertGreaterEqual(obj, min_value, msg)

    def assert_max(self, obj, max_value, msg):
        '''
            Asserts a maximum value for an object i.e. obj <= max_value

            Wrapper on unittest's assertLessEqual

            Args:
                obj: Object of any type which supports <= operator for min_value
                max_value: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertLessEqual(obj, max_value, msg)

    def assert_not_equal(self, obj1, obj2, msg):
        '''
            Assert obj1 != obj2

            Wrapper on unittest's assertNotEqual

            Args:
                obj1: Object of any type which supports != operator for obj2 type
                obj2: Object meeting the above constraint
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertNotEqual(obj1, obj2, msg)

    def assert_true(self, obj, msg):
        '''
            Assert obj is True.

            Wrapper on unittest's assertTrue

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertTrue(obj, msg)

    def assert_false(self, obj, msg):
        '''
            Assert obj is False.

            Wrapper on unittest's assertFalse

            Args:
                obj: Object of any type
                msg: A context string explaining why this assertion was done.
        '''
        self.__asserter.assertFalse(obj, msg)

    def fail(self, msg):
        '''
            Raises AssertionError with the provided message.

            Args:
                msg: A context string explaining the failure.
        '''
        self.__asserter.fail(msg)


class AsserterMixIn:
    '''
        Base class to add **asserter** property to any class which inherits from it.
    '''

    def __init__(self):
        # Trick to use assertions outside of a unittest test
        self.__asserter = Asserter()

    @property
    def asserter(self) -> Asserter:
        '''
            Arjuna's Asserter object for executing assertions.
        '''
        return self.__asserter




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from .nsloader import GuiNamespaceLoaderFactory
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
from arjuna.interact.gui.auto.finder._with import With, Locator
from arjuna.tpi.error import GuiLabelNotPresentError

class GuiDef:
    '''
        A GuiDef object is attached to a Gui, which in turn is attached to an automator and hence to a fixed auto context.

        Arjuna does a lazy loading of Gui Definitions. This means a context A GuiDef will lead to loading of all contexts centrally, but use only the one it needs to avoid repeat processing.
    '''

    def __init__(self, name_store, automator, label, def_file_path): # namespace_dir, 
        self.__name_store = name_store
        # self.__namespace_dir = namespace_dir
        self.__app = automator.app
        self.__automator = automator
        self.__config = automator.config
        self.__auto_context = self.config.guiauto_context
        self.__file_def_path = def_file_path
        self.__ns = None
        ns_name = "file_ns::" + self.__file_def_path.lower()
        if name_store.has_namespace(ns_name):
            self.__ns = name_store.get_namespace(ns_name)
        else:
            self.__ns = name_store.load_namespace(
                ns_name, 
                GuiNamespaceLoaderFactory.create_namespace_loader(
                    self.config,
                    self.__file_def_path
            )
        )

        self.__children = []

    @property
    def config(self):
        return self.__config

    @property
    def is_empty(self):
        return seff.__ns.is_empty()

    def get_wmd(self, label):
        return self.__ns.get_meta_data(label, self.__auto_context)

    @property
    def root_element_name(self):
        try:
            return self.__ns.root_element_name
        except KeyError:
            # Defining __root__ is optional for GNS files.
            return None

    @property
    def anchor_element_name(self):
        try:
            return self.__ns.anchor_element_name
        except KeyError:
            # Defining __load__ is optional for GNS files.
            return None

class GuiFactory:

    @classmethod
    def create_appdef_from_dir(cls, name, automator, app_def_dir):
        from arjuna.tpi.constant import ArjunaOption
        considered_path = app_def_dir
        if not os.path.isdir(considered_path):
            gns_dir = automator.config.value(ArjunaOption.GUIAUTO_NAMESPACE_DIR)
            full_path = os.path.join(gns_dir, considered_path)
            considered_path = os.path.abspath(full_path)
            if not os.path.isdir(considered_path):
                raise Exception("Provided root definition path is not a directory: {}".format(app_def_dir))

        app = GuiDef(automator, os.path.join(considered_path, "HomeGuiPage.yaml"), label=name)
        children_dir = os.path.join(considered_path, "children")
        if os.path.isdir(children_dir):
            lfiles = os.listdir(children_dir)
            for f in lfiles:
                cpath = os.path.join(children_dir, f)
                if os.path.isfile(cpath):
                    base_name = os.path.basename(cpath)
                    app.add_child(base_name, cpath)

    @classmethod
    def create_guidef(cls, automator, def_path):
        return GuiDef(automator, def_path)


# Temp back from GuiDef of older implementation

# def __gns_locators_as_with_locators(self, label):
#     wmd = self.__ns.get_meta_data(label, self.__auto_context)
#     out = []
#     for loc in wmd.raw_locators:
#         underscore = loc.ltype.lower().endswith("attr") and "_" or ""
#         wobj = getattr(With, underscore + loc.ltype.lower()) (loc.lvalue)# e.g. getattr(With, "_" + "ID".lower())("abc")
#         out.append(wobj)
#     return out

# def convert_to_with(self, locator):
#     from arjuna.interact.gui.auto.finder._with import With
#     out_list = []
#     impl_with = locator.as_impl_locator()
#     for wobj in self.__gns_locators_as_with_locators(impl_with.wvalue):
#         # underscore = loc.ltype.lower().endswith("attr") and "_" or ""
#         # wobj = getattr(With, underscore + loc.ltype.lower()) (loc.lvalue)# e.g. getattr(With, "_" + "ID".lower())("abc")
#         if locator.named_args:
#             wobj.format(**locator.named_args)
#         out_list.append(wobj)
#     return out_list
# def convert_to_wmd(self, *locators):
#     final_locators = []
#     for raw_locator in locators:
#         if raw_locator.wtype.upper().strip() == "LABEL":
#             wmd = self.get_wmd(raw_locator.wvalue)
#             for loc in wmd.raw_locators:
#                 if not raw_locator.named_args:
#                     final_locators.append(Locator(ltype=loc.ltype, lvalue=loc.lvalue), named_args={})
#                 else:
#                     final_locators.append(Locator(ltype=loc.ltype, lvalue=loc.lvalue), named_args=raw_locator.named_args)
#         else:
#             if not raw_locator.named_args:
#                 final_locators.append(Locator(ltype=raw_locator.wtype, lvalue=raw_locator.wvalue), named_args={})
#             else:
#                 final_locators.append(Locator(ltype=raw_locator.ltype, lvalue=raw_locator.lvalue), named_args=raw_locator.named_args)
#     return GuiWidgetMetaData(final_locators)



from arjuna.tpi.engine.asserter import Asserter

class Info:

    def __init__(self, pytest_request, attrs=None):
        self.__attrs = attrs
        self.__request = pytest_request
        rnode = self.__request.node
        self.__node_name = rnode.name
        if self.__request.scope == "module":
            self.__orig_name = rnode.name
        elif self.__request.scope == "function":
            self.__orig_name = rnode.originalname and rnode.originalname or rnode.name

    def get_qual_name(self, with_params=False):
        # if pytest name has params only then originalname is set else it is None
        if self.__request.scope in {"module", "session"}:
            return self.__node_name
        else:
            name = with_params and self.__node_name or self.__orig_name
            return self.__request.module.__name__ + "." + name

    def get_qual_name_with_data(self):
        qname = self.get_qual_name(with_params=True)
        if self.__request.fixturename:
            return qname + ":" + self.__request.fixturename
        else:
            return qname

    def __getattr__(self, name):
        if name in self.__attrs:
            return self.__attrs[name]
        else:
            try:
                return getattr(self.__request, name)
            except:
                raise Exception("{name} is not a valid test information attribute. Built-in Arjuna attributes: {attrs}".format(name=name, attrs=str(self.__attrs)))

_LOOKUP_ORDER = {
    "session" : ("session", ),
    "module" : ("session", "module"),
    "class" : ("function", "cls", "module"),
    "function" : ("function", "cls", "module", "session")
}

_SCOPE_MAP = {
    "function" : "function",
    "class"    : "cls",
    "module"   : "module",
    "session"  : "session"
}

class Space:

    def __init__(self, pytest_request):
        vars(self)['_request'] = pytest_request
        try:
            config = self.arj_config
        except:
            from arjuna import Arjuna
            self.arj_config = Arjuna.get_config()

    def __getitem__(self, name):
        scopes = _LOOKUP_ORDER[self._request.scope]
        from arjuna import Arjuna
        for scope in scopes:
            Arjuna.get_logger().debug("Space: Getting value for {} from {} scope".format(name, scope))
            try:
                container = getattr(self._request, _SCOPE_MAP[scope])
                return getattr(container, name)
            except Exception as e:
                Arjuna.get_logger().debug("Space: No value for {} in {} scope".format(name, scope))
                continue
        raise Exception("Attribute with name >>{}<< does not exist in request scope for {}".format(name, scopes))

    def _get_container_for_scope(self):
        return getattr(self._request, _SCOPE_MAP[self._request.scope])

    def __setitem__(self, name, value):
        container = self._get_container_for_scope()
        setattr(container, name, value)

    def __getattr__(self, name):
        if type(name) is str and not name.startswith("__"):
            return self[name]

    def __setattr__(self, name, value):
        container = self._get_container_for_scope()
        from arjuna import Arjuna
        Arjuna.get_logger().debug("Space: Setting {}={} in {} scope".format(name, value, self._request.scope))
        setattr(container, name, value)

    @property
    def raw_request(self):
        return self._request

class GroupSpace(Space):

    def __init__(self, pytest_request):
        super().__init__(pytest_request)

    def _get_container_for_scope(self):
        return getattr(self._request, "session") # Each Arjuna group represents a Pytest session in a thread.

class ModuleSpace(Space):

    def __init__(self, pytest_request):
        super().__init__(pytest_request)

    def _get_container_for_scope(self):
        return getattr(self._request, "module")

class Module:

    def __init__(self, py_request):
        self._space = ModuleSpace(py_request)

    @property
    def space(self):
        return self._space

class Group:

    def __init__(self, py_request):
        self._space = GroupSpace(py_request)
        self._info = py_request.session.group_info

    @property
    def space(self):
        return self._space

    @property
    def thread_name(self):
        return self._info.thread_name

    @property
    def name(self):
        return self._info.name

    @property
    def config(self):
        return self._info.config

class My:

    def __init__(self, test_meta_data=None):
        self._data = None
        self._info = None
        self._handler = None
        self._qual_name = None
        self._request =  None
        self._shared_objects = None
        self._asserter = Asserter() #unittest.TestCase('__init__')
        self._space = None
        self._module = None
        self._attrs = None
        self._tags = None
        self._bugs = None
        self._envs = None
        if test_meta_data:
            self._attrs = test_meta_data['info']
            if self._attrs['id'] is None:
                self._attrs['id'] = self._attrs['qual_name']
            self._tags = test_meta_data['tags']
            self._bugs = test_meta_data['bugs']
            self._envs = test_meta_data['envs']
        self._group = None

    @property
    def config(self):
        return self.space.arj_config

    @property
    def tags(self):
        return self._tags

    @property
    def bugs(self):
        return self._bugs

    @property
    def envs(self):
        return self._envs

    def get_config(self, name=None):
        if name is None:
            return self.config
        else:
            from arjuna import Arjuna
            return Arjuna.get_config(name)
    
    @property
    def data_refs(self):
        from arjuna import Arjuna
        return Arjuna.get_data_references()

    @property
    def group(self):
        '''
            This info is available only within the body of a fixture or test function.
        '''
        # By this stage the Arjuna's built-in default group fixture has executed and group_info is available as pytest_request.session.group_info
        if not self._group:
            self._group = Group(self._request)
        return self._group

    @property
    def module(self):
        return self._module

    @property
    def data(self):
        return self._data

    @property
    def asserter(self):
        return self._asserter

    @data.setter
    def data(self, record):
        self._data = record

    @property
    def space(self):
        return self._space

    def set_req_obj(self, pytest_request):
        self._request = pytest_request
        self._info = Info(pytest_request, self._attrs)
        self._space = Space(pytest_request)
        if pytest_request.scope in {"function"}:
            if not self._module:
                self._module = Module(pytest_request)     

    @property
    def info(self):
        return self._info

    @property
    def resources(self):
        return self._resources

    @property
    def raw_request(self):
        return self._request


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.constant import ArjunaOption
from enum import Enum, auto


class ViewContextHandler:

    def __init__(self, automator):
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def switch_to_view_context(self, view_name):
        self._act(TestAutomatorActionBodyCreator.switch_to_view_context(view_name))

    def get_current_view_context(self):
        response = self._act(TestAutomatorActionBodyCreator.get_current_view_context())
        return response["data"]["viewContext"]

    def get_all_view_contexts(self):
        response = self._act(TestAutomatorActionBodyCreator.get_all_view_contexts())
        return response["data"]["viewContexts"]

    def switch_to_native_view(self):
        self.switch_to_view_context(MobileView.NATIVE_APP.name)

    def switch_to_web_view(self, pkg):
        self.switch_to_view_context(
            "{}_{}".format(MobileView.WEBVIEW.name, pkg)
        )

    def switch_to_first_web_view(self):
        for context_name in self.get_all_view_contexts():
            if context_name.find(MobileView.WEBVIEW.name) != -1:
                self.switch_to_view_context(context_name)
                break

    def _does_name_represent_web_view(self, view_name):
        return view_name.lower().find(MobileView.WEBVIEW.name.lower()) != -1 

    def is_web_view(self):
        return self._does_name_represent_web_view(self.get_current_view_context())

    def is_native_view(self):
        return self.get_current_view_context().lower() == MobileView.WEBVIEW.name.lower()


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import abc
from enum import Enum

from arjuna.tpi.helper.arjtype import CIStringDict
from arjuna.tpi.constant import ArjunaOption
from arjuna.core.yaml import Yaml
from .validator import Validator
from enum import Enum
from arjuna.tpi.constant import BrowserName

class Options(metaclass=abc.ABCMeta):

    def __init__(self, *, options_dict, creation_context, validate=True):
        self.__options = CIStringDict()
        self.__validate = validate
        if options_dict:
            self.update_all(options_dict)

    def update(self, option_name, option_value):
        option_name = self._process_option_name(option_name)
        try:
            is_not_set = False
            try:
                is_not_set = option_value.lower() == "not_set"
            except:
                pass
            finally:
                option_key = self.__get_option_key(option_name)
                if is_not_set:
                    self.__options[option_key] = "not_set"
                else:
                    validator_name, validator = self._get_validator_for(option_name)
                    if self.__validate:
                        self.__options[option_key] = validator(option_value)
                    else:
                        if validator_name.lower() not in {'absolute_dir_path', 'absolute_file_path'}:
                            self.__options[option_key] = validator(option_value)
                        else:
                            self.__options[option_key] = option_value
        except Exception as e:
            raise Exception("Config option value <{}>(type:{}) for <{}> option did not pass the validation check: [{}]".format(
                option_value, type(option_value), option_name, validator_name)
            )
    
    def __get_option_key(self, option_name):
        return isinstance(option_name, Enum) and option_name.name or option_name

    @abc.abstractmethod
    def _process_option_name(self, option_name):
        pass

    @abc.abstractmethod
    def _get_validator_for(self, option_name):
        pass

    def value(self, option_name):
        return self.__options[self.__get_option_key(self._process_option_name(option_name))]

    def update_all(self, options):
        if isinstance(options, Options):
            self.__options.update(options.as_dict())
        else:
            if options:
                for k,v in options.items():
                    self.update(k,v)

    def as_dict(self):
        return self.__options

    def is_not_set(self, option_name):
        self._process_option_name(option_name)
        try:
            return self.value(option_name).upper() == "NOT_SET"
        except:
            return False

class ArjunaOptions(Options):

    ARJUNA_OPTIONS_DESC_MAP = None

    @classmethod
    def process_arjuna_option_name(cls, name):
        try:
            if not isinstance(name, ArjunaOption):
                return ArjunaOption[UserOptions.process_option_name(name)]
            else:
                return name
        except Exception as e:
            raise Exception("Config option <{}> is not a valid ArjunaOption constant".format(name))

    @classmethod
    def load_desc(cls):
        if cls.ARJUNA_OPTIONS_DESC_MAP is not None:
            return
        my_dir = os.path.dirname(os.path.realpath(__file__))
        desc_file = os.path.abspath(os.path.join(my_dir, "..", "res", "arjuna_conf_desc.yaml"))
        creation_context=f"This Yaml represents arjuna_conf_desc.yaml configuration file at {desc_file} that describes rules for Arjuna's built-in options."
        desc_yaml = Yaml.from_file(file_path=desc_file)
        cls.ARJUNA_OPTIONS_DESC_MAP = {cls.process_arjuna_option_name(k): v for k, v in desc_yaml.as_map().items()}

    def __init__(self, *, options_dict, creation_context, validate=True):
        super().__init__(options_dict=options_dict, creation_context=creation_context, validate=validate)

    def _get_validator_for(self, option_name):
        validator_name = self.ARJUNA_OPTIONS_DESC_MAP[option_name]
        return validator_name, getattr(Validator, validator_name.lower())

    def _process_option_name(self, option_name):
        return self.process_arjuna_option_name(option_name)

    def __modify_bin_name_for_windows(self, name):
        if platform.system().lower() == "windows":
            return name + ".exe"
        else:
            return name

    def __get_driver_path(self, name):
        existing_driver_path = self.value(ArjunaOption.SELENIUM_DRIVER_PATH)
        not_set_yet_str = "$driver_name$"
        if existing_driver_path.find(not_set_yet_str) != -1:
            return self.value(ArjunaOption.SELENIUM_DRIVER_PATH).replace(not_set_yet_str, self.__modify_bin_name_for_windows(name))
        else:
            # Some other driver might have been set
            return os.path.join(os.path.dirname(existing_driver_path), name)

    def process_options(self):
        for_browser = {
            BrowserName.CHROME: {
                ArjunaOption.SELENIUM_DRIVER_PROP : "webdriver.chrome.driver",
                ArjunaOption.SELENIUM_DRIVER_PATH : self.__get_driver_path("chromedriver")
            },

            BrowserName.FIREFOX: {
                ArjunaOption.SELENIUM_DRIVER_PROP : "webdriver.gecko.driver",
                ArjunaOption.SELENIUM_DRIVER_PATH : self.__get_driver_path("geckodriver")
            }
        }

        browser = self.get_browser_name()
        self.update_all(for_browser[browser])

    def get_browser_name(self):
        return self.value(ArjunaOption.BROWSER_NAME)

class UserOptions(Options):

    def __init__(self, *, options_dict, creation_context):
        super().__init__(options_dict=options_dict, creation_context=creation_context)

    @classmethod
    def process_option_name(self, name):
        try:
            if isinstance(name, Enum):
                return name.name.upper()
            else:
                return name.upper().strip().replace(".", "_")
        except Exception as e:
            raise Exception("An error occured in processing Config option <{}> as a user option. Error: {}".format(name, str(e)))

    @classmethod
    def pass_through(self, input):
        return input

    def _get_validator_for(self, option_name):
        return "pass_through", self.pass_through

    def _process_option_name(self, option_name):
        return self.process_option_name(option_name)

class EditableConfig:
    _OS_MAP = {
        'Windows': 'windows',
        'Darwin': 'mac',
        'Linux': 'linux'
    }

    def __init__(self, *, arjuna_options_dict, user_options_dict, creation_context, validate=True):
        self.__arjuna_options = ArjunaOptions(options_dict=arjuna_options_dict, creation_context=creation_context, validate=validate)
        self.__user_options = UserOptions(options_dict=user_options_dict, creation_context=creation_context)

    @property
    def arjuna_options(self):
        return self.__arjuna_options

    @property
    def user_options(self):
        return self.__user_options

    def as_dict(self):
        return {
            'arjuna_options': self.arjuna_options.as_dict(),
            'user_options': self.user_options.as_dict()
        }


    def update(self, conf):
        self.__arjuna_options.update_all(conf.arjuna_options)
        self.__user_options.update_all(conf.user_options)

    def update_from_maps(self, *, arjuna_options, user_options):
        self.__arjuna_options.update_all(arjuna_options)
        self.__user_options.update_all(user_options)

    def process_arjuna_options(self):
        self.__arjuna_options.process_options()

    def set_arjuna_option(self, arjuna_option, obj):
        self.__arjuna_options.update(arjuna_option, obj)

    def set_user_option(self, option, obj):
        self.__user_options.update(option, obj)

    def set_option(self, option, obj):
        try:
            self.set_arjuna_option(option, obj)
        except Exception as e:
            self.set_user_option(option, obj)
        return self

    def add_options(self, options):
        for option, obj in options.items():
            self.set_option(option, obj)
        return self

    def is_empty(self):
        if not self.arjuna_options and not self.user_options:
            return True
        return False

    @classmethod
    def empty_conf(cls):
       return EditableConfig(arjuna_options_dict=None, user_options_dict=None, creation_context="")

    @classmethod
    def from_file(cls, *, file_path, creation_context, validate=True, **replacements):
        with open(file_path, "r") as f:
            return cls.from_str(contents=f.read(), creation_context=creation_context, validate=validate, **replacements)

    @classmethod
    def from_str(cls, *, contents, creation_context, validate=True, **replacements):
        for rname, rvalue in replacements.items():
            contents = contents.replace("${}$".format(rname), rvalue)
        
        contents_yaml = Yaml.from_str(name="arjuna_defaults", contents=contents)

        return cls.from_yaml(yaml_obj=contents_yaml, creation_context=creation_context, validate=validate)

    @classmethod
    def from_yaml(cls, *, yaml_obj, creation_context, validate=True):
        return EditableConfig(
            arjuna_options_dict = yaml_obj.get_section("arjuna_options", strict=False).as_map(),
            user_options_dict = yaml_obj.get_section("user_options", strict=False).as_map(),
            creation_context = "This configuration represents " + creation_context,
            validate=validate
        )

    @classmethod
    def arjuna_conf(cls, *, project_root_dir, run_id):
        my_dir = os.path.dirname(os.path.realpath(__file__))
        location = os.path.abspath(os.path.join(my_dir, "..", "res", "arjuna.yaml"))
        return cls.from_file(
            file_path=location,
            creation_context=f"arjuna.yaml default configuration file at {location}",
            arjuna_root_dir = os.path.abspath(os.path.join(my_dir, "..", "..")),
            project_root_dir = project_root_dir,
            project_name = os.path.basename(project_root_dir),
            host_os = cls._OS_MAP[platform.system()],
            run_id = run_id and run_id or "mrun",
            validate=False
        )

    @classmethod
    def project_conf(cls, *, arjuna_conf):
        location = arjuna_conf.arjuna_options.value(ArjunaOption.CONF_PROJECT_FILE)
        proj_conf = cls.empty_conf()
        proj_conf.update(arjuna_conf)
        if not os.path.isfile(location):
            return proj_conf
        proj_conf.update(
            cls.from_file(
                file_path = location,
                creation_context = f"project.yaml configuration file at {location}"
            )
        )
        return proj_conf

    @classmethod
    def from_maps(cls, *, ref_config, arjuna_options, user_options):
        conf = cls.empty_conf()
        if ref_config: conf.update(ref_config)
        conf.update_from_maps(arjuna_options=arjuna_options, user_options=user_options)
        return conf

    @classmethod
    def __multi_confs_file(cls, *, file_path, creation_context):
        data_yaml = Yaml.from_file(file_path=file_path)
        conf_map = dict()
        for section_name in data_yaml.section_names:
            conf_map[section_name] = cls.from_yaml(yaml_obj=data_yaml.get_section(section_name), creation_context=f"Section {section_name} in {creation_context}")

        return conf_map


    @classmethod
    def data_confs(cls, *, arjuna_conf):
        location = arjuna_conf.arjuna_options.value(ArjunaOption.CONF_DATA_FILE)
        if not os.path.isfile(location):
            return {}
        return cls.__multi_confs_file(
            file_path=location, creation_context=f"Data Configuration file at {location}"
        )

    @classmethod
    def env_confs(cls, *, arjuna_conf):
        location = arjuna_conf.arjuna_options.value(ArjunaOption.CONF_ENVS_FILE)
        if not os.path.isfile(location):
            return {}
        return cls.__multi_confs_file(
            file_path=location, creation_context=f"Environment Configuration file at {location}"
        )



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import re

def get_canonical_path(file_path):
    return os.path.abspath(file_path)


def create_file_path(parent_path, child_parts, file_name, extension):
    dir_path = os.pathsep.join([
        parent_path,
        os.pathsep.join(child_parts)
    ])

    os.makedirs(dir_path)
    return os.pathsep.join(dir_path, file_name + extension)


def is_file(path, d=None):
    p = path
    if d:
        p = os.path.join(d, p)
    return os.path.exists(p) and os.path.isfile(p)


def is_dir(path):
    return os.path.exists(path) and os.path.isdir(path)


def is_absolute_path(path):
    return os.path.isabs(path)

def validate_file(path):
    if not os.path.exists(path):
        raise Exception("File does not exist.")
    elif os.path.isfile(path):
        raise Exception("Not a file.")


def validate_dir(path):
    if not os.path.exists(path):
        raise Exception("Directory does not exist.")
    elif os.path.isfile(path):
        raise Exception("Not a directory.")


def get_file_modified_time_stamp(path):
    return os.path.getmtime(path)


def get_latest_file_pathfrom_dir(path):
    check = 0
    last = None
    for f in os.listdir(path):
        m = get_file_modified_time_stamp(os.pathsep.join(path, f))
        if m > check:
            check = m
            last = m

    return last


def delete_dir(dir_path):
    validate_dir(dir_path)
    shutil.rmtree(dir_path)

def delete_dir_if_exists(path):
    if not os.path.exists(path): return
    if not is_dir(path):
        raise Exception("{} is not a directory.".format(path))
    shutil.rmtree(path)

def copy_file(src_file, dest_file):
    validate_file(src_file)
    validate_dir(os.path.dirname(dest_file))
    shutil.copy2(src_file, dest_file)


def copy_file_to_dir(src_file, dest_dir):
    validate_file(src_file)
    validate_dir(dest_dir)
    shutil.copy2(src_file, dest_dir)


def move_file_to_dir(src_file, dest_dir):
    shutil.move(src_file, dest_dir)
    return os.pathsep.join(dest_dir, os.path.basename(src_file))


def create_empty_file(path):
    os.makedirs(os.path.dirname(path))
    f = open("path", "w")
    f.close()


def __base_name_parts(fpath):
    name = os.path.basename(fpath)
    main, *ext = name.rsplit(".", 1)
    if ext:
        return main, ext[0]
    else:
        return main, ""

def get_extension(fpath):
    return __base_name_parts(fpath)[1]

def get_nonext_basename(fpath):
    return __base_name_parts(fpath)[0]

def normalize_path(in_name):
    updated = re.sub(r"[\\]+", "/", in_name)
    updated = re.sub(r"[/]+", "/", updated)
    return updated


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from arjuna.core.reader.excel import *
from arjuna.tpi.engine.data.record import *
from arjuna.tpi.helper.arjtype import CIStringDict

class ContextualDataReference(metaclass=abc.ABCMeta):

    def __init__(self):
        self.map = {}

    def update(self, data_reference):
        for context, record in data_reference.map.items():
            if context not in self.map:
                self.map[context] = CIStringDict()
            self.map[context].update(record.named_values)

    def update_from_dict(self, context, map):
        if context not in self.map:
            self.map[context] = dict()
        self.map[context].update(map)

    def record_for(self, context):
        if context.lower() in self.map:
            return self.map[context.lower()]
        else:
            raise Exception("Context key {} not found in data reference: {}.".format(context, self.__class__.__name__))

    def __str__(self):
        return str({k: str(v) for k,v in self.map.items()})

    def enumerate(self):
        for k,v in self.map.items():
            print(k, "::", type(v), str(v))

class __ExcelDataReference(ContextualDataReference):
    def __init__(self, path):
        super().__init__()
        self.path = path
        if (path.lower().endswith("xls")):
            self.reader = ExcelRow2ArrayReader(path)
        else:
            raise Exception("Unsupported file format for Excel reading.")

        self._populate()

    @abc.abstractmethod
    def _populate(self):
        pass

    def record_for(self, context):
        try:
            return super().record_for(context)
        except:
            raise Exception("{} at {} does not contain {} context key.".format(self.__class__.__name__, self.path, context))


class ExcelRowDataReference(__ExcelDataReference):

    def __init__(self, path):
        super().__init__(path)

    def _populate(self):
        names = self.reader.headers[1:]
        while True:
            try:
                data_record = self.reader.next()
            except StopIteration:
                break
            else:
                self.map[data_record[0].lower()] = DataRecord(context="Ref", **dict(zip(names, data_record[1:])))
        self.reader.close()


class ExcelColumnDataReference(__ExcelDataReference):

    def __init__(self, path):
        super().__init__(path)

    def _populate(self):
        contexts = self.reader.headers[1:]
        cmap = {i: {} for i in contexts}
        while True:
            try:
                data_record = self.reader.next()
            except StopIteration:
                break
            else:
                name = data_record[0]
                for index, context in enumerate(contexts):
                    cmap[context][name] = data_record[index+1]
        self.reader.close()
        for context, kv in cmap.items():
            self.map[context.lower()] = DataRecord(context="Ref", **kv)

def R(query, *, bucket=None, context=None):
    from arjuna import Arjuna, ArjunaOption
    bucket = bucket
    context = context
    query = query

    if bucket is not None:
        bucket = bucket.lower()
        if context is not None:
            context = context.lower()
            query = query.lower()
        else:
            if query.find('.') != -1:
                context, query = query.split('.', 1)
                context = context.lower()
                query = query.lower()
            else:
                raise Exception("DataRefError: The query must specify context using dot notation when not passed as argument.")
    else:
        if context is not None:
            raise Exception("bucket must be provided if context arg is passed.")
        else:
            if query.find('.') != -1:
                bucket, context, query = query.split('.', 2)
                bucket = bucket.lower()
                context = context.lower()
                query = query.lower()
            else:
                raise Exception("DataRefError: The query must specify bucket and context using dot notation when these are not passed as arguments.")           

    try:
        return Arjuna.get_data_ref(bucket).record_for(context)[query]
    except Exception as e:
        import traceback
        raise Exception("Error in retrieving reference value for: bucket: >>{}<<, context: >>{}<< and query >>{}<< in data reference. {}. {}".format(bucket, context, query, str(e), traceback.format_exc()))




# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Classes to assist in XML Parsing. 
'''

from lxml import etree, html
from typing import List, Dict, Tuple

from arjuna.tpi.tracker import track

@track("debug")
class NodeLocator:
    '''
        Locator for finding an XML Node in an **XmlNode**.

        Keyword Arguments:
            tag: (Optional) Tag of the node
            **attrs: Arbitrary number of key value pairs representing attribute name and value.

        Raises:
            Exception: If neither tag nor an attribute is provided.

        Note:
            You can use tag and attributes in combination.

            Supports nested node finding.
    '''
    
    def __init__(self, *, tag: str=None, **attrs):

        if tag is None and not attrs:
            raise Exception("You must provided tag and/or attributes for finding nodes.")

        attr_conditions = []
        if attrs:
            for attr, value in attrs.items():
                if value is None:
                    attr_conditions.append("@{}".format(attr))
                else:
                    attr_conditions.append("contains(@{}, '{}')".format(attr, value))

        attr_str = ""
        if attr_conditions:
            attr_str = "[{}]".format("and ".join(attr_conditions))
        tag = tag and tag or "*"
        prefix = ".//"

        self.__xpath = "{}{}{}".format(prefix, tag, attr_str)

    def search_node(self, xml_node):
        return [XmlNode(n) for n in xml_node.xpath(self.__xpath)]

@track("debug")
class XmlNode:
    '''
        Represents a single node in a parsed XML.

        Arguments:
            node: **lxml** Element object.
    '''

    def __init__(self, node):
        self.__node = node

    @property
    def xml_node(self):
        '''
            Wrapped **lxml** Element

            Not supposed to be used directly.
        '''
        return self.__node

    @property
    def text(self) -> str:
        '''
            Text of the node.

            Note:
                Multiple texts are joined together.
        '''
        return self.xml_node.text

    @property
    def texts(self) -> list:
        '''
            List of Texts of the node.

            Note:
                Multiple texts are stored separately.
        '''
        return self.xml_node.xpath("//text()")

    @property
    def tag(self) -> str:
        '''
            Tag of the node.
        '''     
        return self.xml_node.tag

    @property
    def children(self) -> List['XmlNode']:
        '''
            All Children of this node as a List of XmlNodes
        '''
        return [XmlNode(c) for c in list(self.xml_node)]

    @property
    def parent(self) -> 'XmlNode':
        '''
            Parent XmlNode
        '''
        return XmlNode(self.xml_node.getparent())

    @property
    def preceding_sibling(self) -> 'XmlNode':
        '''
            The XmlNode before this node at same hierarchial level.
        '''
        return XmlNode(self.xml_node.getprevious())

    @property
    def following_sibling(self) -> 'XmlNode':
        '''
            The XmlNode after this node at same hierarchial level.
        '''
        return XmlNode(self.xml_node.getnext())

    @property
    def attrs(self) -> Dict[str,str]:
        '''
            All Attributes of this node as a dictionary.
        '''
        return dict(self.xml_node.attrib)

    def attr(self, name) -> str:
        '''
            Value of an attribute of this node.
        '''
        return self.xml_node.get(name)

    def has_attr(self, name):
        '''
            Check if an attribute is present.
        '''
        return name in self.xml_node.attrib

    def __xpath(self, xpath):
        if not xpath.startswith("."):
            return "." + xpath
        else:
            return xpath

    def findall_with_xpath(self, xpath) -> List['XmlNode']:
        '''
            Find all XmlNodes that match an XPath.
        '''
        return [XmlNode(n) for n in self.xml_node.xpath(self.__xpath(xpath))]

    def find_with_xpath(self, xpath, position=1):
        '''
            Find nth XmlNode that matches an XPath.

            Args:
                xpath: XPath string
                position: XPath index. Default is 1.
        '''
        return self.findall_with_xpath(xpath)[position]

    def as_str(self) -> str:
        '''
            String representation of this node.
        '''
        return etree.tostring(self.xml_node, encoding=str)

    def __str__(self):
        return self.as_str()

    def findall(self, *node_locators, stop_when_matched: bool=False) -> List['XmlNode']:
        '''
            Find all XmlNodes that match one of more `NodeLocator` s.

            Args:
                *node_locators: One or more `NodeLocator` s

            Keyword Arguments:
                stop_when_matched: If True, the call returns nodes found by the first `NodeLocator` that locates one or more nodes. Default is False.

            Returns:
                List of `XmlNode` s. In case of no match, empty list is returned.
        '''

        out = []
        for locator in node_locators:
            nodes = locator.search_node(self.xml_node)
            out.extend(nodes)
            if stop_when_matched:
                if nodes:
                    break
        return out

    def find(self, *node_locators, stop_when_matched: bool=False) -> 'XmlNode':
        '''
            Find first `XmlNode` that match one of more `NodeLocator` s.

            Args:
                *node_locators: One or more `NodeLocator` s

            Keyword Arguments:
                stop_when_matched: If True, the call returns nodes found by the first `NodeLocator` that locates one or more nodes.

            Note:
                In case of no node match, returns None
        '''
        matches = self.findall(*node_locators, stop_when_matched=stop_when_matched)
        if matches:
            return matches[0]
        else:
            return None

    def find_keyvalue_texts(self, key_locator, value_locator) -> Tuple[str, str]:
        '''
            Returns texts of first XmlNodes for a pair of `NodeLocator` s

            Args:
                key_locator: First `NodeLocator` (key)
                value_locator: First `NodeLocator` (value)

            Returns:
                2-element tuple containing the text strings.
        '''
        key = self.find(key_locator).text
        value = self.find(value_locator).text
        return key,value


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Arjuna Types

Contains many general purpose type abstractions.
'''

import abc
import pprint
from collections import OrderedDict
from typing import Callable
import abc

from arjuna.tpi.tracker import track

class _ArDict(metaclass=abc.ABCMeta):
    '''

    '''

    def __init__(self, d=None):
        self.__store = dict()
        if d:
            self.update(d)

    @property
    def store(self):
        return self.__store

    @abc.abstractmethod
    def _process_key(self, key):
        pass

    def __getitem__(self, key):
        return self.__store[self._process_key(key)]

    def pop(self, key):
        return self.__store.pop(self._process_key(key))

    def __setitem__(self, key, value):
        self.__store[self._process_key(key)] = value

    def __delitem__(self, key):
        del self.__store[self._process_key(key)]

    def _update(self, d):
        if not d: return
        for k,v in d.items():
            self[self._process_key(k)] = v

    def update(self, d):
        self._update(d)

    def has_key(self, key):
        return self._process_key(key) in self.__store

    def keys(self):
        return self.__store.keys()

    def __getattr__(self, attr):
        return self.__store[attr]

    def __len__(self):
        return len(self.__store.keys())

    def __str__(self):
        if not self.__store:
            return "<empty>"
        else:
            return str(self.__store)

    def __iter__(self):
        return iter(self.__store)

    def clone(self):
        return self._clone()

    def items(self):
        return self.__store.items()

    def is_empty(self):
        return len(self.__store) == 0


class CIStringDict(_ArDict):
    '''
        Dictionary with case-insensitive keys.

        Arguments:
            pydict: (Optional) A **dict** object.
    '''
    def __init__(self, pydict={}):
        super().__init__(pydict)

    def _process_key(self, key):
        return key.lower()

    def _clone(self):
        return CIStringDict(self.__store)

    def __str__(self):
        return "CIStringDict: " + super().__str__()
        

class ProcessedKeyDict(_ArDict):
    '''
        Dictionary with case-insensitive keys.

        Arguments:
            processor: A callable for processing the dictionary key.
            pydict: (Optional) A **dict** object.
    '''

    def __init__(self, *, processor: Callable, pydict: dict={}):
        self.__processor = processor
        super().__init__(pydict)

    def _process_key(self, key):
        return self.__processor(key)  

    def _clone(self):
        return ProcessedKeyDict(self.__processor, self.__store)  

@track("debug")
class OnceOnlyKeyCIStringDict(CIStringDict):
    '''
        Dictionary with case-insensitive keys that allows for immutable key-value pairs.

        Arguments:
            processor: A callable for processing the dictionary key.
            pydict: (Optional) A **dict** object.
    '''

    def __init__(self, d={}):
        super().__init__(d)

    def __setitem__(self, key, value):
        if self.has_key(key):
            raise Exception("You can not change the value once set.")
        super().__setitem__(key, value)

    def _update(self, d):
        if not d: return
        as_dict = dict(d)
        for k in as_dict:
            self[k] = as_dict.get(k)

    def __iter__(self):
        return super().__iter__()

    def _clone(self):
        return OnceOnlyKeyCIStringDict(self.items())


@track("debug")
class Dictable(metaclass=abc.ABCMeta):
    '''
        Abstract class. Any object which has a method **as_dict** is a **Dictable**.
    '''

    @abc.abstractmethod
    def _as_dict(self):
        pass

    def as_dict(self) -> dict:
        '''
            Dictionary representation of this object.

            Returns:
                A **dict** object.
        '''
        retval = self._as_dict()
        if type(retval) is not dict:
            raise Exception("_as_dict must return a dict type. Got {} of type {}".format(retval, type(retval)))
        return retval

class Point(Dictable):
    '''
        Represents an XY coordinate.

        Args:
            x: X co-ordinate
            y: Y co-ordinate
    '''

    def __init__(self, x, y):
        self.__x = x
        self.__y = y

    @property
    def location(self):
        '''
            Get XY co-ordinates as a tuple -> (x,y)
        '''
        return (self.__x, self.__y)

    def _as_dict(self):
        return {"x": self.__x, "y":self.__y}

class Offset(Point):
    '''
        Represents an offset from current **Point** on **Screen** in terms of XY coordinates.

        Args:
            x: X co-ordinate
            y: Y co-ordinate
    '''

    def __init__(self, x, y):
        super().__init__(x,y)

@track("debug")
class Screen:
    '''
        Represents Gui Screen in terms of its XY coordinates.
    '''

    @staticmethod
    def xy(x:int, y:int) -> Point:
        '''
            Create a **Point** on Gui Screen in terms of its XY coordinates.
        '''
        return _Point(x,y)

    @staticmethod
    def offset(x, y) -> Offset:
        '''
            Create a **Offset** on Gui Screen in terms of its XY coordinates.
        '''
        return _Offset(x,y)

@track("debug")
class NVPair(Dictable):
    '''
        Encapsulates a name-value pair. It is an implementation of **Dictable**.

        Args:
            name: Name of this object
            value: Value of this object
    '''

    def __init__(self, name, value):
        self.__name = name
        self.__value = value

    def _as_dict(self):
        return {"name" : self.__name, "value": self.__value}


@track("debug")
class Attr(NVPair):
    '''
        A name-value pair with an associated optional tag name. It is an implementation of **Dictable**.

        Keyword Arguments:
            name: (Mandatory) Name of this object
            value: (Mandatory) Value of this object
            tag: Tag associated with this object
    '''

    def __init__(self, *, name, value, tag=None):
        super().__init__(name, value)
        self.__tag = tag

    def _as_dict(self):
        d = super()._as_dict()
        d["tag"] = self.__tag
        return d


@track("debug")
class NVPairs(Dictable):
    '''
        Encapsulates arbitrary name-value pairs. It is an implementation of **Dictable**.

        Keyword Arguments:
            **nvpairs: Arbitrary name-value pairs passed as keyword arguments.
    '''

    def __init__(self, **nvpairs):
        self.__kwargs = nvpairs

    def _as_dict(self):
        return self.__kwargs


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from arjuna.tpi.helper.arjtype import _ArDict
from jsonpath_rw import jsonpath, parse
from typing import Any

from arjuna.tpi.tracker import track

@track("debug")
class Json(_ArDict):
    '''
        Arjuna's Json Object.

        Supports dictionary methods as well as **.** access for key.

        Arguments:
            pydict: Python dict.
    '''

    def __init__(self, pydict: dict=None):
        super().__init__(pydict)

    def _process_key(self, key: str):
        return key

    @classmethod
    def from_file(cls, file_path: str) -> 'Json':
        '''
            Creates a Json object from file.

            Arguments:
                file_path: Absolute path of the json file.

            Returns:
                Arjuna's `Json` object
        '''

        with open(file_path, 'r') as f:
            jobj = json.load(f)
            return Json(jobj)

    @classmethod
    def from_map(cls, map):
        '''
            Creates a Json object from Python dictionary.

            Arguments:
                map: Python dict

            Returns:
                Arjuna's `Json` object
        '''

        return Json(d)

    def find(self, query) -> Any:
        '''
            Find element(s) with JsonPath.

            Arguments:
                map: Python dict

            Returns:
                Any object depending on the query.
        '''

        jsonpath_expr = parse(query)
        return [match.value for match in jsonpath_expr.find(self.store)]

    def __getitem__(self, query):
        return self.find(query)

    



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import types
from arjuna.core.utils.obj_utils import get_class_for_method

def singleton(cls):
    @functools.wraps(cls)
    def inner(*args, **kwargs):
        if not hasattr(cls, '_INSTANCE') or cls._INSTANCE is None:
            cls._INSTANCE = cls(*args, **kwargs)
        return cls._INSTANCE

    return inner


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.common.by import By
from selenium.common.exceptions import NoSuchElementException
from arjuna.tpi.error import *
from arjuna.core.error import *

class ElementFinder:
    BY_MAP = {
        "ID": By.ID,
        "NAME": By.NAME,
        "TAG_NAME": By.TAG_NAME,
        "CLASS_NAME": By.CLASS_NAME,
        "LINK_TEXT": By.LINK_TEXT,
        "PARTIAL_LINK_TEXT": By.PARTIAL_LINK_TEXT,
        "CSS_SELECTOR": By.CSS_SELECTOR,
        "XPATH": By.XPATH
    }
    
    @classmethod
    def find_element(cls, container, byType, byValue):
        from arjuna import Arjuna
        Arjuna.get_logger().debug(f"Finding element in container:{container} with wtype:{byType} and wvalue:{byValue}")
        try:
            return container.find_element(cls.BY_MAP[byType.upper()], byValue)
        except Exception as e:
            raise GuiWidgetNotFoundError("By.{}={}".format(byType, byValue))


    @classmethod
    def find_elements(cls, container, byType, byValue):
        elements = container.find_elements(cls.BY_MAP[byType.upper()], byValue)
        if len(elements) == 0:
            raise GuiWidgetNotFoundError("By.{}={}".format(byType, byValue))
        return elements



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Data Generator

This module contains classes and functions that generate data.
'''

import uuid
import random
from mimesis import Person
from mimesis import Address
from mimesis import locales
from mimesis import Text
from collections import namedtuple

Locales = locales
DataEntity = namedtuple

class Random:
    '''
        Provides methods to create random strings and numbers of different kinds.
    '''

    @classmethod
    def ustr(cls, *, prefix: str=None) -> str:
        '''
            Generate a unique UUID string

            Keyword Arguments:
                prefix: (Optional) prefix to be added to the generated UUID string.

            Returns:
                A string that is unique for current session.
        '''
        prefix = prefix and prefix + "-" or ""
        return "{}{}".format(prefix, uuid.uuid4())

    @classmethod
    def first_name(cls, *, locale=Locales.EN):
        '''
            Generate a first name.

            Keyword Arguments:
                locale: (Optional) locale for generating first name

            Returns:
                A generated first name
        '''
        return Person(locale).first_name()

    @classmethod
    def last_name(cls, *, locale=Locales.EN):
        '''
            Generate a last name.

            Keyword Arguments:
                locale: (Optional) locale for generating last name

            Returns:
                A generated last name
        '''
        return Person(locale).last_name()

    @classmethod
    def phone(cls, *, locale=Locales.EN):
        '''
            Generate a phone number.

            Keyword Arguments:
                locale: (Optional) locale for generating phone number

            Returns:
                A generated phone number
        '''
        return Person(locale).telephone()

    @classmethod
    def email(cls, *, locale=Locales.EN):
        '''
            Generate an email address.

            Keyword Arguments:
                locale: (Optional) locale for generating email address

            Returns:
                A generated email address
        '''
        return Person(locale).email()

    @classmethod
    def street_name(cls, *, locale=Locales.EN):
        '''
            Generate a street name

            Keyword Arguments:
                locale: (Optional) locale for generating street name

            Returns:
                A generated street name
        '''
        return Address(locale).street_name()

    @classmethod
    def street_number(cls, *, locale=Locales.EN):
        '''
            Generate a street number

            Keyword Arguments:
                locale: (Optional) locale for generating street number

            Returns:
                A generated street number
        '''
        return Address(locale).street_number()

    @classmethod
    def house_number(cls, *, locale=Locales.EN):
        '''
            Generate a house number

            Keyword Arguments:
                locale: (Optional) locale for generating house number

            Returns:
                A generated house number
        '''
        return cls.street_number(locale=locale)

    @classmethod
    def postal_code(cls, *, locale=Locales.EN):
        '''
            Generate a postal code

            Keyword Arguments:
                locale: (Optional) locale for generating postal code

            Returns:
                A generated postal code
        '''
        return Address(locale).postal_code()

    @classmethod
    def city(cls, *, locale=Locales.EN):
        '''
            Generate a city name.

            Keyword Arguments:
                locale: (Optional) locale for generating city name

            Returns:
                A generated city name
        '''
        return Address(locale).city()

    @classmethod
    def country(cls, *, locale=Locales.EN):
        '''
            Generate a country name

            Keyword Arguments:
                locale: (Optional) locale for generating country name

            Returns:
                A generated country name
        '''
        return Address(locale).country()

    @classmethod
    def sentence(cls, *, locale=Locales.EN):
        '''
            Generate a sentence

            Keyword Arguments:
                locale: (Optional) locale for generating sentence

            Returns:
                A generated sentence
        '''
        return Text(locale).sentence()

    @classmethod
    def fixed_length_number(cls, *, length):
        '''
            Generate a fixed length number

            Keyword Arguments:
                length: Number of digits in generated number.

            Returns:
                A generated fixed length number
        '''
        arr0 = [str(random.randint(1,9))]
        arr = [str(random.randint(0,9)) for i in range(length-1)]
        arr0.extend(arr)
        return "".join(arr0)
        



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import cycle

class PytestHooks:
    '''
        Easy hooks to be used in **pytest** configuration file: **conftest.py** placed under **<Project_Root_Dir/test** directory in test project.
    '''

    @classmethod
    def _get_request_attr(cls, item, obj_name):
        from .test import Space
        request =  cls._get_request_obj(item)
        res = Space(request)
        return getattr(res, obj_name)

    @classmethod
    def _get_request_obj(cls, item):
        return item.funcargs['request']

    @classmethod
    def _get_plugin(cls, item, name):
        return item.config.pluginmanager.getplugin(name)

    @classmethod
    def _get_html_report_plugin(cls, item):
        return cls._get_plugin(item, 'html')

    @classmethod
    def _get_screen_shooter(cls, item):
        try:
            return getattr(item.function, "screen_shooter")
        except:
            try:
                return getattr(item.module, "screen_shooter")
            except:
                return getattr(item.session, "screen_shooter")

    @classmethod
    def add_screenshot_for_result(cls, item, result, *, ignore_passed=True, ignore_fixtures=False):
        '''
            Automatically add screenshot to HTML Report File.

            To be used in **pytest_runtest_makereport** hook in **conftest.py**.

            Args:
                item: **pytest**'s Item object
                result: **pytest**'s TestReport object.

            Keyword Arguments:
                ignore_passed: (Optional) If set to True, screenshot is taken when the test function completes. Default is True.
                ignore_fixtures: (Optional) If set to True, screenshot is not taken for test fixture functions. Default is False.

            Note:
                - For taking the screenshot, it does a look up for a **screen_shooter** attribute in the object spaces in following order:
                    - Function Space
                    - Module Space
                    - Session Space

                - The screen_shooter attribute should contain a **ScreenShooter** i.e. an object of a class that inherits from ScreenShooter class and completes its protocol.
                
                - This is a lenient hook. This means that if any exception happens in it, it ignores the exception and logs a warning message.
        '''

        try:
            try:
                screen_shooter = cls._get_screen_shooter(item)
            except AttributeError:
                return

            html_plugin = cls._get_html_report_plugin(item)
            pytest_html = html_plugin
            report = result.get_result()
            extra = getattr(report, 'extra', [])

            if ignore_fixtures:
                if report.when == 'call':
                    return

            xfail = hasattr(report, 'wasxfail')

            if ignore_passed and report.passed:
                return
            # if (report.skipped and xfail) and (report.failed and not xfail):
                # extra.append(pytest_html.extras.url(app.url))


            import re
            rname = re.sub(r"\[.*?\]", "", report.nodeid)
            image = screen_shooter.take_screenshot(prefix=rname)
            fpath = "../screenshot/{}".format(image.file_name)
            img_elem = '''<img src="data:image/png;base64,{}"/>'''.format(image.base64)
            extra.append(
                pytest_html.extras.html(
                    '''<div class="image"><a href="{}" target="_blank">{}</a>'''.format(fpath, img_elem)
                )
            )
            report.extra = extra
        except Exception as e:
            from arjuna import log_warning
            log_warning("Error in add_screenshot_for_result hook: " + str(e))


    @classmethod
    def configure_group_for_test(cls, metafunc):
        '''
            Configures **group** fixture for a test.

            Acts only if **group** fixture is present in the signature of a test function or signature(s) of any of its fixture(s) in its fixture hierarchy. 

            To be used in **pytest_generate_tests** hook in **conftest.py**.

            Args:
                metafunc: **pytest**'s MetaFunc object

            Note:
                The **group** fixture yields a **DataRecord** object containing the following keys:
                    - **name**: Group name
                    - **config**: **Configuration** object assigned to the group.
                    - **thread_name**: Thread name for the thread in which the Test Group is running.
        '''

        from arjuna import Arjuna, ArjunaOption, log_debug, C
        from arjuna.tpi.engine.data.markup import record
        log_debug("{} {}".format(metafunc.function, metafunc.fixturenames))

        group_params = Arjuna.get_group_params()
        conf = None
        m = metafunc.function.__module__
        group = record(**group_params).build(context='Group').all_records[0]
        log_debug("Parameterizing distributor for module: {} with group: {}".format(m, group))
        metafunc.parametrize("group", argvalues=[group], ids=["G"], indirect=True)

    @classmethod
    def select_tests(cls, pytest_items, pytest_config):
        '''
            Select tests from items collected by pytest, based on Arjuna rules.

            Arguments:
                pytest_items: List of pytest `Item` objects. Each item represents a collected test function node.
                pytest_config: pytest Config object
        '''
        from arjuna import Arjuna
        from arjuna.core.error import ExclusionRuleMet, NoInclusionRuleMet
        selector = Arjuna.get_test_selector()
        final_selection = []
        deselected = []
        for item in pytest_items:
            qual_name = item.nodeid.split('::')[0].replace("/",".").replace('\\',"").replace(".py","") + "." + item.name.split("[")[0]
            try:
                selector.validate(Arjuna.get_test_meta_data(qual_name))
            except (ExclusionRuleMet, NoInclusionRuleMet) as e:
                deselected.append(item)
            else:
                final_selection.append(item)

        if deselected:
            pytest_config.hook.pytest_deselected(items=deselected)
            pytest_items[:] = final_selection


        


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import CIStringDict

from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.utils.repr_utils import repr_dict

class InteractionConfig:

    def __init__(self, meta_dict):
        self.__settings = {
            GuiInteractionConfigType.CHECK_TYPE: True,
            GuiInteractionConfigType.CHECK_PRE_STATE : True,
            GuiInteractionConfigType.CHECK_POST_STATE : True,
            GuiInteractionConfigType.SCROLL_TO_VIEW : False,
        }   

        loaded = []
        for k,v in meta_dict.items():
            try:
                if isinstance(k, GuiInteractionConfigType):
                    self.__settings[k] = v
                else:
                    self._settings[GuiInteractionConfigType[k.upper()]] = v
                loaded.append(k)
            except:
                pass

        for l in loaded:
            del meta_dict[l]

    def should_check_type(self):
        return self.__settings[GuiInteractionConfigType.CHECK_TYPE]

    def should_check_pre_state(self):
        return self.__settings[GuiInteractionConfigType.CHECK_PRE_STATE]

    def should_check_post_state(self):
        return self.__settings[GuiInteractionConfigType.CHECK_POST_STATE]

    def should_scroll_to_view(self):
        return self.__settings[GuiInteractionConfigType.SCROLL_TO_VIEW]

    def __str__(self):
        return repr_dict(self.__settings)

class Meta:

    def __init__(self, mdict=None):
        self.__mdict = not mdict and CIStringDict() or CIStringDict(mdict)
        from arjuna.core.constant import GuiWidgetType
        if "type" in self.__mdict:
            try:
                widget_type = self.__mdict["type"]
                if not isinstance(widget_type, GuiWidgetType):
                    self.__mdict["type"] = GuiWidgetType[widget_type.upper()]
            except Exception as e:
                raise Exception("{} is not a valid Gui widget type.".format(widget_type))
        else:
            self.__mdict["type"] = GuiWidgetType.ELEMENT
        self.__mdict["settings"] = InteractionConfig(self.__mdict) # Interconfig keys are removed

    def update_settings(self, source_wmd):
        self.settings.update(source_wmd.meta.settings)

    def items(self):
        return self.__mdict.items()

    def has(self, name):
        return name.lower() in self.__mdict

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, name):
        if self.has(name):
            return self.__mdict[name]
        else:
            return None

    def __setitem__(self, name, value):
        self.__mdict[name] = value

    def __str__(self):
        return repr_dict(self.__mdict)

    


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import platform

def is_os_windows():
    return platform.system() == "Windows"

def cexit():
	from arjuna import Arjuna
	Arjuna.get_console().display_error("Exiting...")
	sys.exit()

def fexit():
	print("Exiting because of Fatal Error...", file=sys.stderr)
	import time
	time.sleep(0.5)
	sys.exit(1)

# public static Runtime get_run_time() {
#     return Runtime.get_runtime();
# }

# public static String getOSName() {
#     return System.get_property("os.name");
# }

def get_line_separator():
    return os.linesep


def get_path_separator():
    return os.sep



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.common.proxy import *


class BrowserLauncher:

    @classmethod
    def launch(cls, config, svc_url=None):
        driver_path = config["arjunaOptions"]["SELENIUM_DRIVER_PATH"]
        browser_bin_path = config["arjunaOptions"]["BROWSER_BIN_PATH"]
        browser_name = config["arjunaOptions"]["BROWSER_NAME"]
        return CREATOR_MAP[browser_name.name](config, driver_path, browser_bin_path, svc_url)

    @classmethod
    def are_browser_prefs_set(cls, config):
        return "browserPreferences" in config and config["browserPreferences"]

    @classmethod
    def are_browser_args_set(cls, config):
        return "browserArgs" in config and config["browserArgs"]

    @classmethod
    def are_extensions_set(cls, config):
        return "browserExtensions" in config and config["browserExtensions"]

    @classmethod
    def _create_chrome(cls, config, driver_path, browser_bin_path, svc_url):
        from selenium.webdriver import Chrome, ChromeOptions

        caps = DesiredCapabilities.CHROME
        caps.update(config["driverCapabilities"])

        # if config["arjunaOptions"]["BROWSER_PROXY_ON"]:
        #     proxy = Proxy()
        #     proxy_string = "{}.{}".format(
        #         config["arjunaOptions"]["BROWSER_PROXY_HOST"],
        #         config["arjunaOptions"]["BROWSER_PROXY_PORT"]
        #     )
        #     proxy.http_proxy = proxy_string
        #     proxy.ssl_proxy = proxy_string
        #     proxy.add_to_capabilities(caps)

        options = ChromeOptions()

        if browser_bin_path.lower() != "not_set":
            options.binary_location = browser_bin_path

        if cls.are_browser_prefs_set(config):
            options.add_experimental_option("prefs", config["browserPreferences"])

        if config["arjunaOptions"]["BROWSER_HEADLESS"]:
            options.add_argument("--headless")

        if cls.are_browser_args_set(config):
            for arg in config["browserArgs"]:
                options.add_argument(arg)

        if cls.are_extensions_set(config):
            for ext in config["browserExtensions"]:
                options.add_extension(ext)

        caps[ChromeOptions.KEY] = options.to_capabilities()[ChromeOptions.KEY]
        from selenium import webdriver
        return webdriver.Remote(svc_url, caps)

    @classmethod
    def _create_firefox(cls, config, driver_path, browser_bin_path, svc_url):
        from selenium.webdriver import Firefox
        from selenium.webdriver import FirefoxOptions
        from selenium.webdriver import FirefoxProfile

        profile = FirefoxProfile()
        # if config["arjunaOptions"]["BROWSER_PROXY_ON"]:
        #     proxy = Proxy()
        #     proxy_string = "{}.{}".format(
        #         config["arjunaOptions"]["BROWSER_PROXY_HOST"],
        #         config["arjunaOptions"]["BROWSER_PROXY_PORT"]
        #     )
        #     proxy.http_proxy = proxy_string
        #     proxy.ssl_proxy = proxy_string
        #     profile.set_proxy(proxy)

        caps = DesiredCapabilities.FIREFOX
        caps.update(config["driverCapabilities"])

        options = FirefoxOptions()

        if browser_bin_path.lower() != "not_set":
            options.binary_location = browser_bin_path

        if cls.are_browser_prefs_set(config):
            for pref, value in config["browserPreferences"].items():
                options.set_preference(pref, value)

        if config["arjunaOptions"]["BROWSER_HEADLESS"]:
            options.add_argument("-headless")

        if cls.are_browser_args_set(config):
            for arg in config["browserArgs"]:
                options.add_argument(arg)

        from selenium import webdriver
        return webdriver.Remote(svc_url, browser_profile=profile, options=options)

        # driver = Firefox(executable_path=driver_path, firefox_profile=profile, capabilities=caps)
        # if cls.are_extensions_set(config):
        #     for ext in config["browserExtensions"]:
        #         driver.install_addon(ext)

        # return driver

    @classmethod
    def _create_safari(cls, config, driver_path, browser_bin_path):
        from selenium.webdriver import Safari

        caps = DesiredCapabilities.SAFARI
        caps.update(config["driverCapabilities"])

        return Safari(executable_path=driver_path, desired_capabilities=caps)


CREATOR_MAP = {
    "FIREFOX" : BrowserLauncher._create_firefox,
    "CHROME" : BrowserLauncher._create_chrome,
    "SAFARI" : BrowserLauncher._create_safari,
}



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from .enums import WithType, GenericLocateWith
from arjuna.core.utils.repr_utils import repr_dict

class _WithXMetaClass(type):

    def __getattr__(cls, mname):
        from arjuna import Arjuna 
        withx_ref = Arjuna.get_withx_ref()
        if not withx_ref.has_locator(mname):
            raise Exception("There is no built-in or extended locator available as With.{}".format(mname))

        def withx(*vargs, **kwargs): 
            wtype, wvalue = withx_ref.format_args(mname, vargs, kwargs)
            return getattr(With, wtype.lower())(wvalue)
        return withx


class With(metaclass=_WithXMetaClass):

    def __init__(self, with_type, with_value):
        '''

        :param with_type: an enum constant of type WithType
        :param with_value: a string or int or a With object.
        '''
        self.__with_type = with_type
        self.__with_value = None
        self.__has_content_locator = False
        self.__content_locator = None

        if with_type == WithType.ELEMENT:
            if not isinstance(with_value, With):
                raise Exception("For identification with element locator, the argument must be a With object.")
            else:
                self.__has_content_locator = True
                self.__content_locator = with_value
        else:
            self.__with_value = with_value

    @property
    def wtype(self):
        return self.__with_type

    @property
    def wvalue(self):
        return self.__with_value

    @wvalue.setter
    def wvalue(self, value):
        self.__with_value = value

    def as_impl_locator(self):

        impl_with = ImplWith(
            wtype = self.__with_type.name,
            wvalue = self.__has_content_locator and self.__content_locator.as_impl_locator() or self.__with_value,
            has_content_locator = self.__has_content_locator
        )

        return impl_with

    def as_map(self):
        map = dict()
        map["withType"] = self.__with_type.name
        if not self.__has_content_locator:
            map["withValue"] = self.__with_value
        else:
            map["withValue"] = self.__content_locator.as_map()
        
        return map

    @classmethod
    def id(cls, id):
        return With(WithType.ID, id)

    @classmethod
    def name(cls, name):
        return With(WithType.NAME, name)

    @classmethod
    def classes(cls, *names):
        return With(WithType.CLASSES, names)

    @classmethod
    def link(cls, text):
        return With(WithType.LINK, text)

    @classmethod
    def flink(cls, text):
        return With(WithType.FLINK, text)

    @classmethod
    def selector(cls, selector):
        return With(WithType.SELECTOR, selector)

    @classmethod
    def tag(cls, selector):
        return With(WithType.TAG, selector)

    @classmethod
    def xpath(cls, xpath):
        return With(WithType.XPATH, xpath)

    @classmethod
    def text(cls, text, tag=None):
        return With(WithType.TEXT, text)

    @classmethod
    def ftext(cls, text, tag=None):
        return With(WithType.FTEXT, text)

    @classmethod
    def btext(cls, text, tag=None):
        return With(WithType.BTEXT, text)

    @classmethod
    def etext(cls, text, tag=None):
        return With(WithType.ETEXT, text)

    @classmethod
    def title(cls, title, tag=None):
        return With(WithType.TITLE, title)

    @classmethod
    def __validate_attr_map(self, map):
        try:
            rmap = {k.lower():v for k,v in map.items()}
            rmap["name"]
            rmap["value"]
            return rmap
        except:
            raise Exception("Name and value must be supplied. Got: {}".format(map))

    @classmethod
    def _attr(cls, map):
        map = cls.__validate_attr_map(map)
        return With(WithType.ATTR, map)

    @classmethod
    def _fattr(cls, map):
        map = cls.__validate_attr_map(map)
        return With(WithType.FATTR, map)

    @classmethod
    def _battr(cls, map):
        map = cls.__validate_attr_map(map)
        return With(WithType.BATTR, map)

    @classmethod
    def _eattr(cls, map):
        map = cls.__validate_attr_map(map)
        return With(WithType.EATTR, map)   

    @classmethod
    def attr(cls, map):
        return cls._attr(map)

    @classmethod
    def fattr(cls, map):
        return cls._fattr(map)

    @classmethod
    def battr(cls, map):
        return cls._battr(map)

    @classmethod
    def eattr(cls, map):
        return cls._eattr(map)

    @classmethod
    def value(cls, value, tag=None):
        return With(WithType.VALUE, value)

    @classmethod
    def point(cls, point):
        return With(WithType.POINT, point)

    @classmethod
    def js(cls, js):
        return With(WithType.JS, js)

    @classmethod
    def index(cls, index):
        return With(WithType.INDEX, index)

    @classmethod
    def window_title(cls, title):
        return With(WithType.WINDOW_TITLE, title)

    @classmethod
    def window_ptitle(cls, ptitle):
        return With(WithType.WINDOW_PTITLE, ptitle)

    @classmethod
    def element(cls, with_obj):
        return With(WithType.ELEMENT, with_obj)

    @classmethod
    def label(cls, name):
        return With(WithType.LABEL, name)

class ImplWith:
    
    def __init__(self, *, wtype, wvalue, has_content_locator):
        self.wtype = wtype
        self.wvalue = wvalue
        self.has_content_locator = has_content_locator

    def __str__(self):
        return "ImplWith: {}".format(vars(self))


class Locator:

    def __init__(self, ltype, lvalue=None):
        self.ltype = ltype
        self.lvalue = lvalue

    def set_value(self, value):
        self.lvalue = value

    def create_formatted_locator(self, **fargs):

        def get_global_value(in_str):
            from arjuna import C, L, R
            gtype, query = in_str.split(".", 1)
            gtype = gtype.upper()
            return locals()[gtype](query)


        pattern = r"\$(\s*[\w\.]*?\s*)\$"
        repl_dict = {k.lower():v for k,v in fargs.items()}

        new_locator = Locator(self.ltype)

        if self.ltype == GenericLocateWith.ELEMENT:
            new_locator.set_value(locator.lvalue.create_formatted_locator(**fargs))
            return new_locator

        fmt_locator_value = self.lvalue.replace("\{", "__LB__").replace("}", "__RB__")

        # Find params
        matches = re.findall(pattern, fmt_locator_value)
        
        for match in matches:
            names_set = None
            target = "${}$".format(match)
            processed_name = match.lower().strip()
            repl_value = None
            if processed_name.startswith("c.") or processed_name.startswith("l.") or processed_name.startswith("r."):
                repl_value = get_global_value(processed_name)
            else:
                if processed_name not in repl_dict:
                    raise Exception("You must provide a named argument for a custom format string.")
                repl_value = repl_dict[processed_name]

            fmt_locator_value = fmt_locator_value.replace(target, repl_value)
            

        fmt_locator_value = fmt_locator_value.replace("__LB__", "{").replace("__RB__", "}")
        new_locator.set_value(fmt_locator_value)
        return new_locator

    def __str__(self):
        return repr_dict(vars(self), replace_value_enum=True)

    def is_layered_locator(self):
        return False

class GuiGenericLocator(Locator):

    def __init__(self, ltype, lvalue=None):
        super().__init__(ltype, lvalue)

    def as_map(self):
        map = dict()
        map["withType"] = self.ltype.name
        if self.ltype.name.lower() != "content_locator":
            map["withValue"] = self.lvalue
        else:
            map["withValue"] = [l.as_map() for l in self.lvalue.locators]

        # if self.__named_args:
        #     map["named_args"] = self.__named_args
        
        return map

# class GuiGenericChildLocator(Locator):
#     def __init__(self, ltype, lvalue):
#         super().__init__(ltype, lvalue)
    
#     def set_value(self, value):
#         self.lvalue = value    

#     def is_layered_locator(self):
#         return True



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from arjuna.tpi.constant import *
from arjuna.core.constant import *
from arjuna.core.utils import file_utils
from arjuna.core.utils import sys_utils

class DiscoveredFile:
    def __init__(self):
        self.props = {}
        
    def attr(self, attr):
        return self.props[attr]

    def set_attr(self, name, attr):
        self.props[name] = attr


fa_msg1 = '''Duplicate test file found with name: %s.
Check package and class names in test directory.
Arjuna follows a case-INSENSITIVE approach for test names.'''


class FileAggregator:
    def __init__(self):
        self.files = []
        self.found_class_names = set()
        self.temp_map = {}
        from arjuna import Arjuna
        self.logger = Arjuna.get_logger()
        self.console = Arjuna.get_console()

    def add(self, df):
        key = df.attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH) + "/" + df.attr(
            DiscoveredFileAttributeEnum.FULL_NAME).upper()
        if key in self.found_class_names or key in self.temp_map:
            fa_msg1.format(DiscoveredFileAttributeEnum.FULL_NAME)
            self.console.display_error(fa_msg1)
            sys_utils.fexit()
        self.temp_map[key] = df

    def freeze(self):
        paths = []
        paths.extend(self.temp_map.keys())
        paths.sort()
        for path in paths:
            self.files.append(self.temp_map[path])

    def __iter__(self):
        return iter(self.files)

    def enumerate(self):
        for f in self:
            self.logger.debug("-------------------------")
            self.logger.debug("Name:\t" + f.attr(DiscoveredFileAttributeEnum.NAME))
            self.logger.debug("Extension:\t" + f.attr(DiscoveredFileAttributeEnum.EXTENSION))
            self.logger.debug("Full Name:\t" + f.attr(DiscoveredFileAttributeEnum.FULL_NAME))
            self.logger.debug("Package Dot Notation:\t" + f.attr(DiscoveredFileAttributeEnum.PACKAGE_DOT_NOTATION))
            self.logger.debug(
                "Directory Relative Path:\t" + f.attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH))
            self.logger.debug(
                "Directory Absolute Path:\t" + f.attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH))
            self.logger.debug("Comma Separated Relative Path:\t"
                         + f.attr(DiscoveredFileAttributeEnum.COMMA_SEPATARED_RELATIVE_PATH))
            # self.logger.debug("Container:\t" + f.attr(DiscoveredFileAttributeEnum.CONTAINER))
            # self.logger.debug("Container Type:\t" + f.attr(DiscoveredFileAttributeEnum.CONTAINER_TYPE))
            self.logger.debug("-------------------------")


class FileDiscoverer:
    def __init__(self, aggregator, root_dir, include_prefixes=None):
        self.aggregator = aggregator
        if root_dir.endswith("\\") or root_dir.endswith("//"):
            self.root_dir = root_dir[0:-1]
        else:
            self.root_dir = root_dir
        self.root_dir = file_utils.normalize_path(self.root_dir)
        self.cdir = None
        self.cabsdir = None
        self.prefixes = [file_utils.normalize_path(p) for p in include_prefixes]

    def discover(self):
        for d, subdlist, flist in os.walk(self.root_dir):
            normalized_d = file_utils.normalize_path(d)
            if flist:
                for f in flist:
                    full_path = file_utils.normalize_path(os.path.abspath(os.path.join(normalized_d, f)))
                    consider = False
                    for prefix in self.prefixes:
                        if normalized_d.startswith(prefix):
                            consider = True
                            break
                    if not consider: continue
                    file_ext = file_utils.get_extension(full_path)
                    if file_ext.lower() not in set(['py']): continue
                    parent_dir = file_utils.normalize_path(os.path.dirname(full_path))
                    pkg_parent_dir = None  # os.path.commonpath()
                    if parent_dir == self.root_dir:
                        pkg_parent_dir = ""
                    else:
                        pkg_parent_dir = parent_dir[parent_dir.index(self.root_dir) + len(self.root_dir) + 1:]
                    file_ext = file_utils.get_extension(full_path)

                    df = DiscoveredFile()
                    df.set_attr(DiscoveredFileAttributeEnum.NAME, file_utils.get_nonext_basename(full_path))
                    df.set_attr(DiscoveredFileAttributeEnum.EXTENSION, file_ext)
                    df.set_attr(DiscoveredFileAttributeEnum.FULL_NAME, f)
                    df.set_attr(DiscoveredFileAttributeEnum.PACKAGE_DOT_NOTATION,
                                 pkg_parent_dir.replace("/", "."))
                    df.set_attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH, pkg_parent_dir)
                    df.set_attr(DiscoveredFileAttributeEnum.DIRECTORY_ABSOLUTE_PATH, parent_dir)
                    replaced = df.attr(DiscoveredFileAttributeEnum.DIRECTORY_RELATIVE_PATH).replace("/", "|")
                    replaced = replaced.replace("\\", "|")
                    df.set_attr(DiscoveredFileAttributeEnum.COMMA_SEPATARED_RELATIVE_PATH,
                                 ",".join(replaced.split("|")))
                    # df.set_attr(DiscoveredFileAttributeEnum.CONTAINER, attr.NA_STRING)
                    # df.set_attr(DiscoveredFileAttributeEnum.CONTAINER_TYPE, attr.NA_STRING)
                    self.aggregator.add(df)

        self.aggregator.freeze()



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.tracker import track

@track("debug")
class GuiSourceContent:
    '''
        Source content of a Gui or GuiWidget.

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_entity.source.content

        Arguments:
            all: Complete XML source
            root: Source of XML root.
            inner: All source minus the root source.
            text: Text content.
    '''

    def __init__(self, all, root, inner, text):
        self.__all = all
        self.__root = root
        self.__inner = inner
        self.__text = text

    @property
    def all(self) -> str:
        '''
            Complete source of associated Gui entity.
        '''
        return self.__all

    @property
    def root(self) -> str:
        '''
            Source of root node of associated Gui entity.
        '''
        return self.__root

    @property
    def inner(self) -> str:
        '''
            Complete source of associated Gui entity minus the root node source.
        '''
        return self.__inner

    @property
    def text(self) -> str:
        '''
            All text content of associated Gui entity.
        '''
        return self.__text


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.constant import ArjunaOption

class WebAlert:

    def __init__(self, automator):
        self.__automator = automator

    @property
    def automator(self):
        return self.__automator

    def confirm(self):
        self.__automator.dispatcher.confirm_web_alert()
        self.__automator.alert_handler.delete_alert()

    def dismiss(self):
        self.__automator.dispatcher.dismiss_web_alert()
        self.__automator.alert_handler.delete_alert()

    @property
    def text(self):
        return self.__automator.dispatcher.get_text_from_web_alert()

    @text.setter
    def text(self, text):
        self.__automator.dispatcher.send_text_to_web_alert(text)


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.interact.gui.auto.finder.wmd import SimpleGuiWidgetMetaData
from arjuna.tpi.guiauto.meta.locator import GuiWidgetLocator

from arjuna.tpi.guiauto.source.element import GuiElementSource
from arjuna.tpi.tracker import track

@track("debug")
class GuiDropDown:
    '''
        Represents a drop down list in th Gui.

        Not meant to be directly created. It is created using calls from **Gui** object or **GuiNamespace** object of **Gui**.

        Arguments:
            gui: **Gui** object containing this GuiDropDown.
            wmd: **GuiElementMetaData** object for this GuiDropDown.

        Keyword Arguments:
            parent: **GuiElement** in case it is found inside a **GuiElement**. Default is the **Gui** object.
    '''

    def __init__(self, gui, wmd, parent=None):
        self.__gui = gui
        self.__wmd = wmd
        self.__automator = gui._automator
        self.__finder = parent and parent or gui
        self._wrapped_main_element = self.__finder._wmd_finder.element(wmd)
        self.__found = False
        self.__options = None
        self.__option_locator = self._wmd.meta.option_locator is not None and self._wmd.meta.option_locator or GuiWidgetLocator(type="multi_element", tag="option")

        # It is seen in some websites like Bootstrap based that both select and options are children of a main div element.
        self.__option_container_same_as_select = self._wmd.meta.option_container_locator is None and True or False
        if not self.__option_container_same_as_select:
            self.__option_container = self.__finder.element(self._wmd.meta.option_container_locator)

        self.__source_parser = None

        self.__find()

    @property
    def _wmd(self):
        return self.__wmd

    @property
    def gui(self) -> 'Gui':
        '''
            **Gui** object containing this GuiDropDown.
        '''
        return self.__gui
        
    def __validate_select_control(self, tag):
        if tag.lower() != "select":
            raise Exception("The element should have a 'select' tag for WebSelect element. Found: " + tag)
        self._multi = self.__is_multi_select()

    def __find(self):

        def check_type_if_configured(tag):
            if self.__wmd.meta.settings.should_check_type(): self.__validate_select_control(tag)

        def get_root_element():
            return self.__option_container_same_as_select and self._wrapped_main_element or self.__option_container

        def load_options():
            container = get_root_element()
            self.__options = container.locate(self.__option_locator)

        # self._wrapped_main_element.find()
        tag = self._wrapped_main_element.source.tag
        check_type_if_configured(tag)
        load_options()

    def __is_multi_select(self):
        source = self._wrapped_main_element.source
        return source.get_attr_value("multiple", optional=True) is True or source.get_attr_value("multi", optional=True) is True

    def is_multi_select(self) -> bool:
        '''
            Check if this GuiDropDown allows multiple selection of options.
        '''
        return self._multi

    def has_index_selected(self, index: int) -> bool:
        '''
            Check if this GuiDropDown has option selected at given index.

            Args:
                index: Target index.
        '''
        return self.__options[index].is_selected()

    def has_value_selected(self, value: str) -> bool:
        '''
            Check if this GuiDropDown has option with given value attribute content selected.

            Args:
                value: Exact content of value attribute.
        '''
        return self.__options.get_element_by_value(value).is_selected()

    def has_visible_text_selected(self, text: str) -> bool:
        '''
            Check if this GuiDropDown has option with visible text selected.

            Args:
                text: Exact visible text content.
        '''
        return self.__options.get_element_by_visible_text(text).is_selected()

    def __select_option(self, option):
        self._wrapped_main_element.click()
        option.select()
        if self.__wmd.meta.settings.should_check_post_state() and not option.is_selected():
            raise Exception("The attempt to select the dropdown option was not successful.")

    def select_index(self, index: int) -> None:
        '''
            Select option at given index.

            Args:
                index: Target index.
        '''
        option = self.__options[index]
        self.__select_option(option)

    def select_ordinal(self, ordinal: int) -> None:
        '''
            Select option at given ordinal.

            Ordinals are as per human counting. First element is at ordinal 1.

            Args:
                ordinal: Target ordinal.
        '''
        return self.select_by_index(ordinal-1)

    def select_text(self, text: str) -> None:
        '''
            Select option with given visible text.

            Args:
                text: Exact visible text content.
        '''
        option = self.__options.get_element_by_visible_text(text)
        self.__select_option(option)

    @property
    def text(self) -> str:
        '''
            Visible text of selected option

           Note:
           
            You can select an option with visible text of this GuiDropDown too using following Python code:
                .. code-block:: python

                    dropdown.text = "<some_text>"

            **Method used for this diffent from **select_text****. It uses **send_keys** to simulate this interaction instead of clicking an option.
            
            Waits for clickability.

            **ArjunaOption.GUIAUTO_MAX_WAIT** in associated configuration is used as the default. Can be overriden using **max_wait** argument in GuiWidgetLocator or GNS file.
        '''
        option = self.__options.first_selected_element
        return option.text

    @text.setter
    def text(self, text):
        # Dropdown element does not support clear text.
        self._wrapped_main_element.enter_text(text)

    def select_value(self, value: str) -> None:
        '''
            Select option with given content of value attribute.

            Args:
                value: Exact content of value attribute.
        '''
        option = self.__options.get_element_by_value(value)
        self.__select_option(option)

    @property
    def value(self) -> str:
        '''
            Content of value attribute of selected option
        '''
        option = self.__options.first_selected_element
        return option.source.get_attr_value("value")

    @property
    def source(self) -> GuiElementSource:
        '''
            **GuiSource** for this GuiDropDown (source of root element).
        '''
        return self._wrapped_main_element.source

    def __validate_multi_select(self):
        if not self.is_multi_select():
            raise Exception("Deselect actions are allowed only for a multi-select dropdown.")

    def _deselect_by_value(self, value):
        self.__validate_multi_select()
        return self.__options.get_element_by_value(value).deselect()

    def _deselect_by_index(self, index):
        pass

    def _deselect_by_visible_text(self, text):
        pass

    def _get_selected_options(self):
        pass

    def _are_visible_texts_selected(self, text_list):
        pass

    def _are_values_selected(self, text_list):
        pass

    def _all_options(self):
        pass

    def _select_by_values(self, value_list):
        pass

    def _deselect_by_values(self, value_list):
        pass

    def _select_by_indices(self, indices):
        pass

    def _deselect_by_indices(self, indices):
        pass

    def _select_by_visible_texts(self, text_list):
        pass

    def _deselect_by_visible_texts(self, text_list):
        pass


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .gui import Gui
from .section import GuiSection
from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData

class GuiDialog(GuiSection):
    '''
        Represents a GUI Dialog. Can be associated with a **GuiApp** or **GuiPage**.

        It is an implementation of **GuiAppContent**.

        Args:
            *args: Any number of positional argumnts. These are passed to the **prepare()** method if defined in inherited class.

        Keyword Arguments:
            parent_gui: (Mandatory) The **Gui** object that is associated with this **GuiDialog**.
            label: Label for the this **GuiDialog**. If not provided, the class name is used as the label.
            root: Root element of this **GuiSection**. Can be **label** string defined in its GNS File or a **Locator** object.
            gns_dir: Relative Root Directory for GNS file associated with this **GuiDialog**. Default is **dialog** directory in associated **GuiApp** namespace. If provided, path is considered relative to the **GuiApp** namespace directory.
            gns_file_name: Name of GNS file associated with this **GuiPage**. If not provided, default is **<label>.yaml**.
            kwargs: Arbitrary keyword arugments. These are passed to the **prepare()** method if defined in inherited class.

        Note:
            A **GuiDialog** can have a root element. If defined, all locator calls in this object happen as a nested locating call using the root element.

                - Root element provided in in __init__ call is given preference.
                - If not provided, Arjuna looks for **root** definition in **load** section of its GNS file.
                - If above is also not provided, then locating happens from the root of the DOM of current page.
    '''

    def __init__(self, *args, parent_gui: 'GuiAppOrGuiPage', label: 'str'=None, root: 'LabelOrLocator'=None, gns_dir: str=None, gns_file_name:str=None, **kwargs):
        super().__init__(*args, parent_gui=parent_gui, label=label, root=root, gns_dir=gns_dir, gns_file_name=gns_file_name, **kwargs)






# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from arjuna.tpi.helper.arjtype import Dictable
from arjuna.tpi.tracker import track

@track("debug")
class GuiWidgetLocator(Dictable):
    '''
        Representation of indentifiers, format arguments, widget configuration and named arguments for a GuiWidget.

        Keyword Arguments:
            type: type of GuiWidget. Default is 'element'/GuiWidgetType.ELEMENT. 
            fmt_args: Dictionary of key-value pairs to format the identifiers.
            **named_args: Named arguments representing allowed identifiers, extended identifiers, configuration options and GuiWidget arguments.
    '''

    def __init__(self, *, type="element", fmt_args=None, **named_args):
        self.__widget_type = type

        self.__fmt_args = fmt_args
        if self.__fmt_args is None:
            self.__fmt_args = dict()

        self.__named_args = named_args

        from arjuna.interact.gui.auto.finder.meta import Meta
        self.__meta = Meta({
            "type" : type,
        })

    @property
    def widget_type(self):
        '''
            Type of this GuiWidget.
        '''
        return self.__widget_type

    @property
    def fmt_args(self):
        '''
            Dictionary of key-value pairs to format the identifiers.
        '''
        return self.__fmt_args

    @property
    def named_args(self):
        '''
            Named arguments representing allowed identifiers, extended identifiers, configuration options and GuiWidget arguments.
        '''
        return self.__named_args

    def __str__(self):
        return str(self.as_dict())

    def _as_dict(self):
        return {
                "type": self.widget_type,
                "fmt_args": self.fmt_args,
                "meta": str(self.__meta)
            }

    def _as_raw_wmd(self):
        from arjuna import Arjuna
        from arjuna.interact.gui.auto.finder._with import With
        from arjuna.interact.gui.auto.finder.enums import WithType
        with_list = []
        for k,v in self.__named_args.items():
            if k.upper() in WithType.__members__:
                if isinstance(v, Dictable):
                    v = v.as_dict()
                with_list.append(getattr(With, k.lower())(v))
            elif Arjuna.get_withx_ref().has_locator(k):
                if isinstance(v, Dictable):
                    v = v.as_dict()
                    with_list.append(getattr(With, k.lower())(**v))
                else:
                    with_list.append(getattr(With, k.lower())(v))
            else:
                self.__meta[k] = v
        if not with_list:
            raise Exception("You must provide atleast one locator.")
        from arjuna.interact.gui.auto.finder.wmd import GuiWidgetMetaData
        return GuiWidgetMetaData.create_wmd(*with_list, meta=self.__meta)

    def _as_wmd(self):
        '''
            Convert this **GuiWidgetLocator** to **GuiWidgetMetaData** object
        '''
        wmd = self._as_raw_wmd()
        fmt_wmd = wmd.create_formatted_wmd(**self.__fmt_args)
        return fmt_wmd


# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import copy
import abc
from lxml import etree, html
from io import StringIO
from collections import namedtuple
from arjuna.tpi.helper.xml import XmlNode
from arjuna.tpi.tracker import track
from .content import GuiSourceContent


def _process_child_html(in_str):
    processed = os.linesep.join([l for l in in_str.splitlines() if l.strip()])
    return "\t" + processed

def _remove_empty_lines_from_string(in_str):
    return ' '.join([l.strip() for l in in_str.splitlines() if l.strip()])

def _empty_or_none(in_str):
    if type(in_str) is str and not in_str.strip():
        return True
    else:
        return in_str is None

@track("debug")
class GuiSource:
    '''
        GUI Source of a GUI entity.

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_entity.source
    '''
    def __init__(self):
        self.__content = None

    @property
    def content(self) -> GuiSourceContent:
        '''
            Source content of associated GUI entity returned as :class:`~arjuna.tpi.guiauto.source.content.GuiSourceContent`.
        '''
        return self.__content

    @content.setter
    def _content(self, content):
        self.__content = content

@track("debug")
class SingleGuiEntitySource(GuiSource, metaclass=abc.ABCMeta):
    '''
        Abstract Base class for GUI Source of a Singular GUI entity (**GuiPage** or **GuiElement**).

        Not meant to be directly constructed by a test author. Retrieve as:
        
            .. code-block:: python
                
                gui_entity.source

        Args:
            raw_source: Raw XML source.
            root_tag: 'html' for Gui and 'body' for GuiElement.
    '''

    def __init__(self, raw_source, root_tag):
        super().__init__()
        self.__raw_source = raw_source
        self.__root_tag = root_tag
        self.__fpaths = []
        self.__node = None 
        self.__elem_node = None

    @property
    def node(self) -> XmlNode:
        '''
            Source code as an Arjuna :class:`~arjuna.tpi.helper.xml.XmlNode` for advanced inquiry and parsing.
        '''
        return self.__node

    @property
    def _elem_node(self):
        return self.__elem_node

    def _process_elem_node(self, elem_node):
        pass

    def _load(self):
        raw_source = self.__raw_source
        parser = etree.HTMLParser(remove_comments=True)
        tree = etree.parse(StringIO(raw_source), parser)
        # Done separately from above to retain all original content
        self.__node = XmlNode(etree.parse(StringIO(raw_source), parser))
        
        if self.__root_tag == "body":
            body = tree.getroot().find('body')
            elem_node = list(body)[0]
        else:
            body = tree.getroot()
            elem_node = body

        self.__elem_node = elem_node

        normalized_text_content = os.linesep.join(
            [
                _remove_empty_lines_from_string(c.text)
                for c in elem_node.iter() 
                if not _empty_or_none(c.text)
            ]).strip()

        normalized_inner_html = os.linesep.join([
            _process_child_html(etree.tostring(c, encoding='unicode'))
            for c in list(elem_node.iterchildren())
            ])

        self.__text_content = normalized_text_content
        self.__inner_html = normalized_inner_html
        self.__full_source = raw_source
        for child in list(elem_node): elem_node.remove(child)
        self.__self_source = ' '.join(etree.tostring(elem_node, encoding=str).splitlines())
        self.__self_source = re.sub(r"\s+", " ", self.__self_source)

        self._content = GuiSourceContent(all=self.__full_source, root=self.__self_source, inner=self.__inner_html, text=self.__text_content)

        self._process_elem_node(elem_node)



# This file is a part of Arjuna
# Copyright 2015-2020 Rahul Verma

# Website: www.RahulVerma.net

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from .source import *
from .reference import *
from arjuna.core.utils import file_utils
from arjuna.core.constant import DataRefContextType

def get_data_file_path(data_dir, fpath):
    if file_utils.is_absolute_path(fpath):
        if not file_utils.is_file(fpath):
            if file_utils.is_dir(fpath):
                raise Exception("Not a file: {}".format(fpath))
            else:
                raise Exception("File does not exist: {}".format(fpath))
        return fpath
    else:
        fpath = os.path.abspath(os.path.join(data_dir, fpath))
        if not file_utils.is_file(fpath):
            if file_utils.is_dir(fpath):
                raise Exception("Not a file: {}".format(fpath))
            else:
                raise Exception("File does not exist: {}".format(fpath))
        return fpath

def create_file_data_source(file_path, *, context, delimiter="\t"):
    from arjuna import Arjuna, ArjunaOption
    data_dir = Arjuna.get_config().value(ArjunaOption.DATA_SRC_DIR)
    file_path = get_data_file_path(data_dir, file_path)
    ds = None
    ext = file_path.lower()
    if ext.endswith(".csv") or ext.endswith(".txt"):
        ds = DsvFileMapDataSource(file_path, delimiter, context=context)
    elif ext.endswith(".xls"):
        ds = ExcelFileMapDataSource(file_path, context=context)
    elif ext.endswith(".ini"):
        ds = IniFileDataSource(file_path, context=context)
    else:
        raise Exception("This is not a default file format supported as a data source: " + path)
    return ds

from arjuna.tpi.helper.arjtype import CIStringDict

class DataReferences:

    def __init__(self):
        vars(self)['_store'] = CIStringDict()

    def __getitem__(self, name):
        return self._store[name]

    def __setitem__(self, name, value):
        self._store[name] = value

    def __getattr__(self, name):
        if type(name) is str and not name.startswith("__"):
            return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __str__(self):
        return str(self._store)

class DataReference:

    @classmethod
    def load_all(cls, ref_config):
        from arjuna.tpi.constant import ArjunaOption
        refs = DataReferences()
        column_data_ref_dir = ref_config.value(ArjunaOption.DATA_REF_COLUMN_DIR)
        if os.path.isdir(column_data_ref_dir):
            for fname in os.listdir(column_data_ref_dir):
                if fname.lower().endswith("xls"):
                    refs[os.path.splitext(fname)[0]] = DataReference.create_excel_column_data_ref(fname)
        
        row_data_ref_dir = ref_config.value(ArjunaOption.DATA_REF_ROW_DIR)
        if os.path.isdir(row_data_ref_dir):
            for fname in os.listdir(row_data_ref_dir):
                if fname.lower().endswith("xls"):
                    refs[os.path.splitext(fname)[0]] = DataReference.create_excel_row_data_ref(fname)

        return refs

    @classmethod
    def __create_excel_file_data_ref(cls, file_path, context=DataRefContextType.COLUMN):
        ext = file_path.lower()
        if not ext.endswith("xls"):
            raise Exception("Unsupported file extension for Excel data reference: {}".format(file_path))

        from arjuna import Arjuna, ArjunaOption
        if context == DataRefContextType.COLUMN:
            data_dir = Arjuna.get_config().value(ArjunaOption.DATA_REF_COLUMN_DIR)
            file_path = get_data_file_path(data_dir, file_path)
            return ExcelColumnDataReference(file_path)
        elif context == DataRefContextType.ROW:
            data_dir = Arjuna.get_config().value(ArjunaOption.DATA_REF_ROW_DIR)
            file_path = get_data_file_path(data_dir, file_path)
            return ExcelRowDataReference(file_path)

    @classmethod
    def create_excel_column_data_ref(cls, file_path):
        return cls.__create_excel_file_data_ref(file_path, context=DataRefContextType.COLUMN)

    @classmethod
    def create_excel_row_data_ref(cls, file_path):
        return cls.__create_excel_file_data_ref(file_path, context=DataRefContextType.ROW)

